program display;

{*****************************************************************************************
 *
 * Project   : display
 * Unit Name : display
 * Purpose   : show or change display settings (resolution, frequency, bit depth)
 * Author    : markus stephany
 * Copyright : (C) markus stephany, www.mirkes.de, merkes@mirkes.de

      This package is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your option)
      any later version.

      This package is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with this package in the file named "COPYING"; if not,
      write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
      Boston, MA 02111-1307, USA.

 * Version   : 0.1, 11/15/2003
 * History   : 0.1 - 11/15/2003 : initial release
 *           : 09-27-2004: compiled using delphi 7
 *
 *****************************************************************************************}


{$APPTYPE CONSOLE}

uses
  Windows;

{$I ..\LIB\VERSION.INC}

const
  DISP_CHANGE_BADDUALVIEW =    -6;

(******************* ErrWrite:
  write text to stderr *)
procedure ErrWrite(sMsg: string);
var
  dwDummy: Cardinal;
begin
  if sMsg <> ''
  then
    WriteFile(GetStdHandle(STD_ERROR_HANDLE), sMsg[1], Length(sMsg), dwDummy, nil);
end;

(******************* ErrWriteLn:
  put a line to stderr *)
procedure ErrWriteLn(sMsg: string);
begin
  ErrWrite(sMsg+#13#10);
end;

(******************* Bail:
 print an error message to stdout and exit *)
procedure Bail(sMsg: string; iExit: Integer=1);
begin
  ErrWriteLn(sMsg);
  Halt(iExit);
end;

(******************* StripPath:
  strip the path and ext from the given file name *)
function StripPath(sFile: string): string;
begin
  Result := sFile;
  while Pos('\',Result) > 0
  do
    Delete(Result,1,Pos('\',Result));
  if Copy(Result,Length(Result)-3,1) = '.'
  then
    SetLength(Result, Length(Result)-4);
end;

(******************* Usage:
  print usage message *)
procedure Usage(sErrOpt:string = '');
var
  sApp: string;
begin
  sApp := StripPath(ParamStr(0));
  // possibly print the unknown option
  if sErrOpt <> '' then
  begin
    ErrWriteln(sApp+': '+sErrOpt+'.');
    ErrWriteln('');
  end;
  ErrWriteln(sApp+' v'+BT_VERSION+' (C) markus stephany (GPL)');
  ErrWriteln('      show or change display settings (resolution, frequency, bit depth)');
  ErrWriteln('usage:');
  ErrWriteln(sApp+' h<width> v<height> [f<freq.>] [b<bit depth>]: change display settings');
  ErrWriteln(StringOfChar(' ', Length(sApp))+
                  ' (e.g. h1024 v768 f60 b16)');
  ErrWriteln(sApp+' /LIST: list all supported screen resolutions');
  ErrWriteln(sApp+' /CUR: show current display settings');
        Bail(sApp+' /?: shows this usage screen');
end;


(******************* a2i:
  string to integer routine *)
function a2i(const s: string; var i: integer): boolean;
var
  c: integer;
begin
  Val(s,i,c);
  Result := c = 0;
end;

(******************* ia:
  integer to string routine *)
function i2a(const i: integer): string;
begin
  Str(i, Result);
end;

(******************* up:
  convert to upper case characters *)
function Up(const sIn: string): string;
var
  i: integer;
begin
  Result := '';
  if Length(sIn) > 0 then
    for i := 1 to Length(sIn) do
      Result := Result + UpCase(sIn[i]);
end;

(******************* ListPossibleResolutions:
  display all supported resolutions *)
procedure ListPossibleResolutions;
var
  i: integer;
  mode: TDevMode;
begin
  FillChar(mode, sizeof(mode), #0);
  i := 0;
  mode.dmSize := sizeof(mode);
  while EnumDisplaySettings(nil, i, mode) do
  begin
    writeln('h'+i2a(mode.dmPelsWidth)+' v'+i2a(mode.dmPelsHeight)+' f'+i2a(mode.dmDisplayFrequency)+
      ' b'+i2a(mode.dmBitsPerPel));
    inc(i);
  end;
  if i = 0 then
    Bail('Cannot enumerate supported display settings.')
  else
    Halt(0);
end;

(******************* DisplayCurrentResolution:
  display the current resolution *)
procedure DisplayCurrentResolution;
var
  DC: HDC;
  h,v,f,b: integer;
begin
  DC := GetDC(0);
  h := GetDeviceCaps(DC,HORZRES);
  v := GetDeviceCaps(DC,VERTRES);
  b := GetDeviceCaps(DC,BITSPIXEL);
  f := GetDeviceCaps(DC,VREFRESH);
  ReleaseDC(0, DC);
  writeln('h'+i2a(h)+' v'+i2a(v)+' f'+i2a(f)+' b'+i2a(b));
  Halt(0);
end;

var
  bErr: boolean;
  sErr: string;
  i: integer;

  h,v,b,f: integer;

  mode: TDevMode;
begin
  // check if special parameters are on the command line
  if ParamStr(1) = '/?' then
    Usage;
  if Up(ParamStr(1)) = '/LIST' then
    ListPossibleResolutions;
  if Up(ParamStr(1)) = '/CUR' then
    DisplayCurrentResolution;
  // scan command line
  sErr := '';
  bErr := ParamCount = 0;
  if not bErr then
  begin
    h := -1;
    v := -1;
    b := -1;
    f := -1;
    for i := 1 to ParamCount do
    begin
      sErr := ParamStr(i);
      case UpCase(sErr[1]) of
        'H': if not a2i(Copy(sErr, 2, MaxInt), h) then
        begin
          sErr := 'Invalid parameter value: '+sErr;
          bErr := True;
          Break;
        end;
        'V': if not a2i(Copy(sErr, 2, MaxInt), v) then
        begin
          sErr := 'Invalid parameter value: '+sErr;
          bErr := True;
          Break;
        end;
        'B': if not a2i(Copy(sErr, 2, MaxInt), b) then
        begin
          sErr := 'Invalid parameter value: '+sErr;
          bErr := True;
          Break;
        end;
        'F': if not a2i(Copy(sErr, 2, MaxInt), f) then
        begin
          sErr := 'Invalid parameter value: '+sErr;
          bErr := True;
          Break;
        end;
      else
        sErr := 'Unrecognized parameter: '+sErr;
        bErr := True;
        Break;
      end;
    end;
  end
  else
    Usage(sErr);

  if not bErr then
  begin
    if v = -1 then
    begin
      bErr := True;
      sErr := 'Missing v<height> parameter';
    end;
    if h = -1 then
    begin
      bErr := True;
      sErr := 'Missing h<width> parameter';
    end;
    if bErr then
      Usage(sErr);
  end;

  if not bErr then
  begin
    FillChar(Mode, sizeof(mode), #0);
    Mode.dmSize := sizeof(Mode);
    if h <> -1 then
    begin
      mode.dmPelsWidth := h;
      mode.dmFields := mode.dmFields or DM_PELSWIDTH;
    end;

    if v <> -1 then
    begin
      mode.dmPelsHeight := v;
      mode.dmFields := mode.dmFields or DM_PELSHEIGHT;
    end;

    if b <> -1 then
    begin
      mode.dmBitsPerPel := b;
      mode.dmFields := mode.dmFields or DM_BITSPERPEL;
    end;

    if f <> -1 then
    begin
      mode.dmDisplayFrequency := f;
      mode.dmFields := mode.dmFields or DM_DISPLAYFREQUENCY;
    end;

    i := ChangeDisplaySettings(mode, CDS_TEST);
    if i = DISP_CHANGE_SUCCESSFUL then
      i := ChangeDisplaySettings(mode, 0);

    bErr := True;
    case i of
      DISP_CHANGE_BADDUALVIEW: sErr := 'The settings change was unsuccessful because system is DualView capable.';
      DISP_CHANGE_BADFLAGS: sErr := 'An invalid set of flags was passed in.';
      DISP_CHANGE_BADMODE: sErr := 'The graphics mode is not supported.';
      DISP_CHANGE_BADPARAM: sErr := 'An invalid parameter was passed in. This can include an invalid flag or combination of flags.';
      DISP_CHANGE_FAILED: sErr := 'The display driver failed the specified graphics mode.';
      DISP_CHANGE_NOTUPDATED: sErr := 'Unable to write settings to the registry.';
      DISP_CHANGE_RESTART: sErr := 'The computer must be restarted in order for the graphics mode to work.';
      DISP_CHANGE_SUCCESSFUL: bErr := False;
    else
      sErr := 'An unknown error occured while changing the display settings.';
    end;

  end;

  if not bErr then
  begin
    writeln('Display settings successfully changed.');
    Halt(0)
  end
  else
    Bail(sErr);
end.
