program filesize;

(*

  returns the 32 bit part of the file size given on the command line as exit code,
  DWORD(-1), if no such file has been found.

  if the option /r <value> is specified,
  0 is returned if the file size equals <value>
  1 is returned if the file size is less than <value>
  2 is returned if the file size is greater than <value>
  3 is returned if the file cannot be read

  if the options /r <value> /s are specified
  DWORD(-1) is returned if the file size is less than <value>
  0 is returned if the file size equals <value>
  1 is returned if the file size is greater than <value>
  3 is returned if the file cannot be read

  (C) markus stephany, released under the terms of the GNU GPL version 2.

*)

{$APPTYPE CONSOLE}
uses
  Windows, SysUtils;

{$I ..\LIB\VERSION.INC}

(******************* ErrWrite:
  write text to stderr *)

procedure ErrWrite(sMsg: string);
var
  dwDummy: Cardinal;
begin
  if sMsg <> '' then
    WriteFile(GetStdHandle(STD_ERROR_HANDLE), sMsg[1], Length(sMsg), dwDummy,
      nil);
end;

(******************* ErrWriteLn:
  put a line to stderr *)

procedure ErrWriteLn(sMsg: string);
begin
  ErrWrite(sMsg + #13#10);
end;

(******************* Bail:
 print an error message to stdout and exit *)

procedure Bail(sMsg: string; iExit: Integer = 1);
begin
  ErrWriteLn(sMsg);
  Halt(iExit);
end;

(******************* Usage:
  print usage message *)

procedure Usage(sErrOpt: string = '');
var
  sApp: string;
begin
  sApp := ChangeFileExt(ExtractFileName(ParamStr(0)), '');
  // print the unknown option
  if sErrOpt <> '' then
  begin
    ErrWriteln(sApp + ': unknown option "' + sErrOpt + '".');
    ErrWriteln('');
  end;
  ErrWriteln(sApp + ' version ' + BT_VERSION + ' (C) markus stephany (GPL)');
  ErrWriteln('      returns a file''s size as errorlevel or compares it to a reference value');
  ErrWriteln('usage:');
  ErrWriteln(sApp + ' <filename> = returns <filename>''s file size as');
  ErrWriteln('      errorlevel (lower 32 bits), -1 is returned on error');
  ErrWriteln(sApp +
    ' <filename> /r <size reference> = compares <filename>''s size to');
  ErrWriteln('      <size reference> and returns one of the following errorlevels:');
  ErrWriteln('      0: file size equals reference value');
  ErrWriteln('      1: file size is greater than reference value');
  ErrWriteln('      2: file size is less than reference value');
  ErrWriteln('      3: an error occurred');
  ErrWriteln(sApp +
    ' <filename> /r <size reference> /s = compares <filename>''s size');
  ErrWriteln('      to <size reference> and returns one of the following errorlevels:');
  ErrWriteln('     -1: file size is less than reference value');
  ErrWriteln('      0: file size equals reference value');
  ErrWriteln('      1: file size is greater than reference value');
  ErrWriteln('      3: an error occurred');
  Bail(sApp + ' /? = shows this usage screen');
end;

(******************* CheckParam:
 check wheter paramstr(iIndex) is one of the specified options *)

function CheckParam(iIndex: Integer; const aChars: TSysCharSet): Boolean;
begin
  Result := (Length(ParamStr(iIndex)) = 2) and
    (ParamStr(iIndex)[1] in ['/', '-']) and
    (ParamStr(iIndex)[2] in aChars);
end;

(******************* PrintVersion:
  print program version to stdout and exit *)

procedure PrintVersion;
begin
  writeln(ExtractFileName(ParamStr(0)), ' version ', BT_VERSION);
  Halt(0);
end;

var
  sFileName: string; // first parameter = file name to retrieve filesize
  bSigned: boolean; // signed error levels?
  iReference: int64; // reference value
  bUseRef: boolean; // use reference value?
  d: TWin32FindData;
  h: THandle;
  i, c: integer;
  iSize: int64; // actual size
begin
  sFileName := '';
  bSigned := False;
  bUseRef := False;
  i := 1;
  iReference := 0;

  (* check the command line *)
  while ParamCount >= i do
  begin
    if CheckParam(i, ['?', 'H', 'h']) then
      Usage
    else if CheckParam(i, ['V']) then
      PrintVersion
    else
    begin
      if i = 1 then
        sFileName := ParamStr(i)
      else
      begin
        if CheckParam(i, ['s']) then
          bSigned := True
        else if CheckParam(i, ['r']) then
        begin
          if i = ParamCount then
            Usage;
          bUseRef := True;
          Inc(i);
          Val(ParamStr(i), iReference, c);
          if c <> 0 then
            Usage;
        end
        else
          Usage(ParamStr(i));
      end;
    end;
    Inc(i);
  end;

  if (bSigned and (not bUseRef)) or (sFileName = '') then
    Usage;

  h := FindFirstFile(PChar(sFileName), d);
  if h <> INVALID_HANDLE_VALUE then
  begin
    Windows.FindClose(h);
    iSize := (Int64(d.nFileSizeHigh) shl 32) or (d.nFileSizeLow);
    if not bUseRef then
      ExitProcess(iSize)
    else
    begin
      if iSize = iReference then
        ExitProcess(0)
      else if iSize > iReference then
        ExitProcess(1)
      else
      begin
        if bSigned then
          ExitProcess(Cardinal(-1))
        else
          ExitProcess(2)
      end;
    end;
  end
  else
  begin
    if bUseRef then
      ExitProcess(3)
    else
      ExitProcess(Cardinal(-1));
  end;
end.

