{*****************************************************************************************
 *
 * Project   : flushdrv
 * Unit Name : flushdrv
 * Purpose   : flush specified (on command line) or all local drives / only nt
 * Author    : markus stephany
 * Copyright : (C) markus stephany, www.mirkes.de, merkes@mirkes.de

      This package is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your option)
      any later version.

      This package is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with this package in the file named "COPYING"; if not,
      write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
      Boston, MA 02111-1307, USA.

 * Version   : 0.021, 11/15/2003
 * History   : 0.01 - 22.05.01 : initial release
 *             0.02 - 23.05.01 : added /m option (mounts on windows 2000)
 *             0.021- 15.11.03 : changed some defs from const to var
 *             09-27-2004 : compiled using delphi 7
 *
 *****************************************************************************************}

{*****************************************************************************************
 *
 * Acknowledgement:

    this software is derived from the GPL c program "sync" written by
    Ahdrey Shedel <andreys@cr.cyco.com>, many thanks!

    differences:

     - written in delphi
     - more than one drive can be specified on the command line
     - it uses win32 api calls (CreateFile, FlushFileBuffers) rather than native api
       (NtOpenFile, NtFlushBuffersFile); this seems to make no difference !?
     - it can also flush volumes mounted to a subfolder of another drive (ntfs5)


 *
 *****************************************************************************************}


program flushdrv;

uses
  Windows;

{$I ..\LIB\VERSION.INC}

var
  b_Mounts : Boolean = False; // /m mounted devices option
  hModKernel : THandle = 0; // dynamic kernel32.dll handle

(* nt 5 specific *)
var
  FindFirstVolumeMountPoint: function( lpszRootPathName: PChar;       // volume name
                                       lpszVolumeMountPoint: PChar;   // output buffer
                                       cchBufferLength: DWORD         // size of output buffer
                                      ): THandle; stdcall;

  FindNextVolumeMountPoint: function( hFindVolumeMountPoint: THandle; // search handle
                                      lpszVolumeMountPoint: PChar;    // output buffer
                                      cchBufferLength: DWORD          // size of output buffer
                                     ): BOOL; stdcall;

  FindVolumeMountPointClose: function( hFindVolumeMountPoint: THandle // search handle
                                      ): BOOL; stdcall;

  GetVolumeNameForVolumeMountPoint: function ( lpszVolumeMountPoint: PChar; // volume mount point or directory
                                               lpszVolumeName: PChar;       // volume name buffer
                                               cchBufferLength: DWORD       // size of volume name buffer
                                              ): BOOL; stdcall;

(******************* ErrWrite:
  write text to stderr *)
procedure ErrWrite(sMsg: string);
var
  dwDummy: Cardinal;
begin
  if sMsg <> ''
  then
    WriteFile(GetStdHandle(STD_ERROR_HANDLE), sMsg[1], Length(sMsg), dwDummy, nil);
end;

(******************* ErrWriteLn:
  put a line to stderr *)
procedure ErrWriteLn(sMsg: string);
begin
  ErrWrite(sMsg+#13#10);
end;

(******************* Bail:
 print an error message to stdout and exit *)
procedure Bail(sMsg: string; iExit: Integer=1);
begin
  ErrWriteLn(sMsg);
  if hModKernel <> 0
  then
    FreeLibrary(hModKernel);
  hModKernel := 0;
  Halt(iExit);
end;

(******************* StripPath:
  strip the path and ext from the given file name *)
function StripPath(sFile: string): string;
begin
  Result := sFile;
  while Pos('\',Result) > 0
  do
    Delete(Result,1,Pos('\',Result));
  if Copy(Result,Length(Result)-3,1) = '.'
  then
    SetLength(Result, Length(Result)-4);
end;

(******************* Usage:
  print usage message *)
procedure Usage(sErrOpt:string = '');
var
  sApp: string;
begin
  sApp := StripPath(ParamStr(0));
  // eventually print the unknown option
  if sErrOpt <> '' then
  begin
    ErrWriteln(sApp+': unknown option "'+sErrOpt+'".');
    ErrWriteln('');
  end;
  ErrWriteln(sApp+' version '+BT_VERSION+' (C) markus stephany (GPL)');
  ErrWriteln('      flush local drive buffers on windows nt');
  ErrWriteln('usage:');
  ErrWriteln(sApp+' [/m]                   flushes all local drives');
  ErrWriteln(sApp+' X: [Y:...] [/m]        flushes only the specified drives');
  ErrWriteln(StringOfChar(' ',Length(sApp))+
                  ' /m                     searches for mounted devices on ntfs5'#13#10);
  ErrWriteln(sApp+' /V                     prints the program''s version info');
        Bail(sApp+' /?                     shows this usage screen');
end;

(******************* IsWinNT:
 returns True if the os is a windows nt *)
function IsWinNT(cMinVer:Cardinal=4): Boolean;
var
  rOsInfo: TOSVersionInfo;
begin
  Result := False;
  ZeroMemory(@rOsInfo,sizeof(rOsInfo));
  rOsInfo.dwOSVersionInfoSize := sizeof(rOsInfo);
  if GetVersionEx(rOsInfo)
  then
    Result := rOsInfo.dwPlatformId = VER_PLATFORM_WIN32_NT;
  Result := Result and (rOsInfo.dwMajorVersion >= cMinVer);
end;

// set of chars for option parsing
type
  TCharSet = set of Char;

(******************* CheckParam:
 check wheter paramstr(iIndex) is one of the specified options *)
function CheckParam(iIndex: Integer; const aChars: TCharSet): Boolean;
begin
  Result := (Length(ParamStr(iIndex)) = 2) and
            (ParamStr(iIndex)[1] in ['/','-']) and
            (ParamStr(iIndex)[2] in aChars);
end;

(******************* PrintVersion:
  print program version to stdout and exit *)
procedure PrintVersion;
begin
  writeln(StripPath(ParamStr(0)),' version ',BT_VERSION);
  if hModKernel <> 0
  then
    FreeLibrary(hModKernel);
  hModKernel := 0;
  Halt(0);
end;

(******************* IsValidDriveString:
  is the option a valid drive letter (+ colon) *)
function IsValidDriveString(sIn: string): Boolean;
begin
  Result := ((Length(sIn) = 1) or ((Length(sIn) = 2) and (sIn[2] = ':'))) and
            (UpCase(sIn[1]) in ['A'..'Z']);
end;

(******************* GetLastErrorMessage:
 get the system's last error message *)
function GetLastErrorMessage: string;
var
  sIn: string;
begin
  SetLength(sIn, 1024);
  SetLength(sIn, FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_ARGUMENT_ARRAY,
                               nil,
                               GetLastError,
                               0,
                               @sIn[1],
                               1023,
                               nil));
  while (sIn <> '') and (sIn[Length(sIn)] < #33)
  do
    SetLength(sIn,Pred(Length(sIn)));

  // since this is a console app, convert to oem
  SetLength(Result, Length(sIn));
  if sIn <> ''
  then
    CharToOemBuff(PChar(sIn), PChar(Result),Length(sIn));

end;

(******************* HandleFlushError:
  error while syncing one drive *)
procedure HandleFlushError(sMsg: string);
begin
  if sMsg = ''
  then
    sMsg := GetLastErrorMessage;
  ErrWriteln('failed: '+sMsg);
end;


(******************* DeleteTrailingBackSlash:
  delete trailing back slashes *)
function DeleteTrailingBackslash(sIn: string): string;
begin
  Result := sIn;
  while Copy(Result,Length(Result),1) = '\'
  do
    Delete(Result,Length(Result),1);
end;

(******************* AddTrailingBackslash:
  add a trailing back slash *)
function AddTrailingBackslash(sIn: string): string;
begin
  Result := DeleteTrailingBackslash(sIn)+'\';
end;

(******************* IsNTFS:
 check if ntfs *)
function IsNTFS(sIn: string): Boolean;
var
  szBuf: array[0..10] of char;
  dwMax, dwFlags: Cardinal;
begin
  Result := False;
  if GetVolumeInformation(PChar(AddTrailingBackslash(sIn)), nil, 0, nil, dwMax, dwFlags, szBuf, sizeof(szBuf))
  then
    Result := szBuf = 'NTFS';
end;

(******************* Flush1Drive:
  flush the file buffers of one drive *)
procedure Flush1Drive(sDrive: string; sDev:string='');
var
  hDev: THandle;
  szBuf, szBuf1: array[0..MAX_PATH] of Char;
begin
  if sDev = ''
  then
    sDev := sDrive;

  // handle mounted devices
  if b_Mounts and IsNTFS(sDrive) then
  begin
    hDev := FindFirstVolumeMountPoint(PChar(AddTrailingBackslash(sDrive)), szBuf, sizeof(szBuf));
    if hDev <> INVALID_HANDLE_VALUE then
    begin
      repeat
        if not GetVolumeNameForVolumeMountPoint(PChar(AddTrailingBackslash(sDrive)+szBuf),szBuf1, sizeof(szBuf1)) then
        begin
          ErrWrite(sDrive+' mount points: ');
          HandleFlushError('');
          CloseHandle(hDev);
          Break;
        end;

        Flush1Drive(AddTrailingBackslash(sDrive)+szBuf, Copy(szBuf1,5,MaxInt)); // delete \\?\
      until not FindNextVolumeMountPoint(hDev, szBuf, sizeof(szBuf));
      FindVolumeMountPointClose(hDev);
    end
    else
      if GetLastError <> ERROR_NO_MORE_FILES then
      begin
        ErrWrite(sDrive+' mount points: ');
        HandleFlushError('');
      end;
  end;

  write(sDrive,' ');
  if sDev <> sDrive
  then
    write('(\',sDev,') ');

  if GetDriveType(PChar(sDrive)) <> DRIVE_FIXED then
  begin
    HandleFlushError('not a fixed drive');
    Exit;
  end;

  hDev := CreateFile(PChar('\\.\'+DeleteTrailingBackslash(sDev)),
                     GENERIC_READ or GENERIC_WRITE,
                     FILE_SHARE_READ or FILE_SHARE_WRITE,
                     nil,
                     OPEN_EXISTING,
                     0,// FILE_FLAG_WRITE_THROUGH or FILE_FLAG_NO_BUFFERING, (* seems to be not needed *)
                     0);

  if hDev = INVALID_HANDLE_VALUE then
  begin
    HandleFlushError('');
    Exit;
  end;

  if not FlushFileBuffers(hDev) then
  begin
    HandleFlushError('');
    CloseHandle(hDev);
    Exit;
  end;

  CloseHandle(hDev);
  Writeln('done');

end;

(******************* FlushAllDrives:
 *)
procedure FlushAllDrives;
var
  sDrives: set of 0..25;
  iNum: Integer;
  sDrive: string;
begin
  writeln('syncing disks...');
  Integer(sDrives) := GetLogicalDrives;
  for iNum := 0 to 25
  do
    if iNum in sDrives then
    begin
      sDrive := Char(Ord('A')+iNum)+':\';
      if GetDriveType(PChar(sDrive)) = DRIVE_FIXED
      then
        Flush1Drive(sDrive);
    end;
end;

(******************* CheckGetProcAddress:
  get proc address or return false *)
function CheckGetProcAddress(var pfunc: Pointer; pFuncName: PChar; hMod: THandle): Boolean;
begin
  pFunc := GetProcAddress(hMod, pFuncName);
  Result := Assigned(pFunc);
end;

(******************* InitMountApi:
 assign addresses to the nt 5 specific api calls *)
function InitMountApi: Boolean;
var
  cErr: Cardinal;
begin
  Result := False;
  hModKernel := LoadLibrary('kernel32.dll');
  if hModKernel <> 0 then
  begin
    Result := CheckGetProcAddress(@FindFirstVolumeMountPoint,'FindFirstVolumeMountPointA',hModKernel) and
              CheckGetProcAddress(@FindNextVolumeMountPoint,'FindNextVolumeMountPointA',hModKernel) and
              CheckGetProcAddress(@FindVolumeMountPointClose,'FindVolumeMountPointClose',hModKernel) and // not ansi/unicode
              CheckGetProcAddress(@GetVolumeNameForVolumeMountPoint,'GetVolumeNameForVolumeMountPointA',hModKernel);

    if not Result then
    begin
      cErr := GetLastError;
      FreeLibrary(hModKernel);
      hModKernel := 0;
      SetLastError(cErr);
    end;
  end;
end;

(******************* :
  void main() *)
var
  i:Integer;
  bHaveDrives: Boolean;
begin
  (* needs windows nt *)
  if not IsWinNT then
  begin
    Bail('i need windows nt to run!');
  end;

  bHaveDrives := False;

  (* check the command line *)
  if ParamCount > 0 then
  begin
    for i := 1 to ParamCount
    do
      if CheckParam(i,['?','H','h'])
      then
        Usage
      else
        if CheckParam(i,['V'])
        then
          PrintVersion
        else
          if CheckParam(i,['m'])
          then
            b_Mounts := True
          else
            if not IsValidDriveString(ParamStr(i))
            then
              Usage (ParamStr(i))
            else
              bHaveDrives := True;
    end;

    // mounts option only available on nt5
    if b_Mounts then
    begin
       if not IsWinNt(5)
       then
         Bail('the /m option needs windows 2000/xp!')
       else
       if not InitMountAPI
       then
         Bail('Cannot initialize mount api functions: '#13#10+GetLastErrorMessage);
    end;

    if bHaveDrives then
    begin
      writeln('syncing disks...');

      for i := 1 to ParamCount
      do
        if IsValidDriveString(ParamStr(i))
        then
          Flush1Drive(Copy(ParamStr(i),1,1)+':\');
    end
    else
      FlushAllDrives;

  if hModKernel <> 0
  then
    FreeLibrary(hModKernel);
  hModKernel := 0;
end.

