program gpscript;

{$APPTYPE CONSOLE}

(*

  gpscript, a command line tool to view and modify windows 2000/xp local group policy
  files (Registry.pol) via script.

  program version: 0.02, alpha stage!

  (C) markus stephany merkes@mirkes.de, released under the GNU GPL

*)

{$DEFINE DEBUG}

uses
  Windows,
  SysUtils,
  Classes,
  StrUtils,
  regpol2kxp,
  utils,
  cl_utils;

resourcestring
  ERR_UNKNOWN_ARG = 'unknown parameter in %s: %s';
  ERR_NOT_COMMAND = 'not a command in %s: %s';
  ERR_ARG_NOT_ALLOWED = 'no argument allowed in %s for %s';
  ERR_ARG_NEEDED = 'argument needed in %s for %s';
  ERR_UNKNOWN_TYPE = 'unknown data type "%s" in %s';
  ERR_ARG_MISSING = 'argument "%s" is missing for command "%s" in %s';
  ERR_INVALID_DWORD = 'invalid /DATA:D argument "%s" in %s';
  ERR_INVALID_BINARY = 'invalid /DATA:B argument "%s" in %s';
  ERR_UNKNOWN_DATA = 'unknown /DATA format "%s" in %s';

var
  DoCreate: Boolean = False; // do not create .pol file, override with /CREATE
  DoUpdateGptIni: Boolean = False;
    // do not update gpt.ini[general]version, override with /INI
  FFileName: string = ''; // name of the .pol file, 1st arg on the command line
{$IFDEF DEBUG}
  DoDebug: Boolean = False;
  // if /DEBUG, don't actually do, but print what would be modified
{$ENDIF}

{$I ..\LIB\VERSION.INC}

const
  GPScriptVersion = 'gpscript '+BT_VERSION;

  // gpext.dll guids
  GPEXT_DLL_MACHINE = '[{35378EAC-683F-11D2-A89A-00C04FBBCFA2}{0F6B957D-509E-11D1-A7CC-0000F87571E3}]';
  GPEXT_DLL_USER    = '[{35378EAC-683F-11D2-A89A-00C04FBBCFA2}{0F6B957E-509E-11D1-A7CC-0000F87571E3}]';

  // known commands / options
const
  KNOWN_CMDS: array[0..{$IFDEF DEBUG}13{$ELSE}12{$ENDIF}] of string = (
    '/KEY', // 0 - option
    '/VALUE', // 1 - option
    '/TYPE', // 2 - option
    '/DATA', // 3 - option
    '/CREATE', // 4 - flag
    '/DUMP', // 5 - command
    '/FILE', // 6 - command
    '/SET', // 7 - command
    '/ADD', // 8 - command
    '/MODIFY', // 9 - command
    '/DELETE', //10 - command
    '/CLEAR', //11 - command
    '/INI' //12 - command
{$IFDEF DEBUG}
    , '/DEBUG' //13 - flag
{$ENDIF}
    );

type
  TArg = (aOKey, aOValue, aOType, aOData, aFCreate, aCDump, aCFile, aCSet,
    aCAdd, aCModify, aCDelete, aCClear, aCGptIni, {$IFDEF DEBUG}aFDebug,
{$ENDIF}aOErrFile);

  TPolType = (ptFile, ptMachine, ptUser);

var
  POL_TYPE: TPolType = ptFile; // assume arbitrary file policy

  // show help

procedure Usage;
begin
  writeln(ErrOutput, GPScriptVersion, ' (', RegPol2kXPVersion, ')');
  writeln(ErrOutput, CLEFT);
  writeln(ErrOutput);
  writeln(ErrOutput,
    'gpscript can change win 2k/xp local group policy settings (registry.pol files)');
  writeln(ErrOutput,
    'based on options and commands from files, stdin or the command line.');
  writeln(ErrOutput);
  writeln(ErrOutput, 'usage:');
  writeln(ErrOutput, 'gpscript <registry.pol> [<options>] <commands>');
  writeln(ErrOutput);
  writeln(ErrOutput, '  <registry.pol>:');
  writeln(ErrOutput,
    '    the location of a local win 2k/xp group policy file to edit');
  writeln(ErrOutput,
    '    (e.g. %systemroot%\system32\group policies\machine\registry.pol).');
  writeln(ErrOutput, '    two filenames have a special meaning:');
  writeln(ErrOutput,
    '    "MACHINE" will be replaced by the name of the local machine policy file,');
  writeln(ErrOutput,
    '    "USER" will be replaced by the name of the local user policy file.');
  writeln(ErrOutput);
  writeln(ErrOutput, '  <options>:');
  writeln(ErrOutput, '    /KEY:"<key name>"');
  writeln(ErrOutput,
    '    specifies the registry key name relative to HKLM or HKCU root');
  writeln(ErrOutput, '    (depending on the location of the registry.pol file).');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /VALUE:"<value name>"');
  writeln(ErrOutput,
    '    specifies the name of the registry value to store in the policy file');
  writeln(ErrOutput,
    '    (an empty value name (/VALUE:"") means the key''s default value).');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /TYPE:<data type>');
  writeln(ErrOutput,
    '    specifies the data type of the value to store in the policy file');
  writeln(ErrOutput,
    '    (either decimal, hexadecimal (prefixed by 0x) or one of the NT4');
  writeln(ErrOutput, '    REG_XX constants).');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /DATA:<data>');
  writeln(ErrOutput, '    the actual data that should be stored in the .pol file.');
  writeln(ErrOutput, '    see "data formats" below for further information.');
  writeln(ErrOutput);
  writeln(ErrOutput, '  <commands>:');
  writeln(ErrOutput, '    /CLEAR');
  writeln(ErrOutput, '    delete all existing entries in the current policy file.');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /CREATE');
  writeln(ErrOutput,
    '    this command tells gpscript to create the <registry.pol> file');
  writeln(ErrOutput, '    (and its path) if it doesn''t yet exist.');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /INI[:<name of the gpt.ini>]');
  writeln(ErrOutput,
    '    this command tells gpscript to update the [general]/version item');
  writeln(ErrOutput,
    '    of the specified ini file. if the name is omitted the local gpt.ini');
  writeln(ErrOutput, '    (%systemroot%\system32\grouppolicy\gpt.ini) is used.');
  writeln(ErrOutput, '    additionally, gpext.dll is registered in the ini file.');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /DUMP');
  writeln(ErrOutput,
    '    dump the contents of the whole registry.pol file to stdout in a format');
  writeln(ErrOutput, '    that may be used by gpscript (see the /FILE command).');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /FILE[:<name of command file>]');
  writeln(ErrOutput,
    '    read options/commands from the file <name of command file>.');
  writeln(ErrOutput,
    '    if the command file name is omitted, read options/commands from STDIN.');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /SET');
  writeln(ErrOutput,
    '    sets the data (whose properties are defined by the <options>)');
  writeln(ErrOutput,
    '    if the key/value already exists, the data will be modified.');
  writeln(ErrOutput,
    '    if the key/value doesn''t exist, the data will be added.');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /ADD');
  writeln(ErrOutput,
    '    adds the data (whose properties are defined by the <options>)');
  writeln(ErrOutput, '    the key/value must NOT already exist in the database.');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /MODIFY');
  writeln(ErrOutput,
    '    modifies the data (whose properties are defined by the <options>)');
  writeln(ErrOutput, '    the key/value MUST already exist in the database.');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /DELETE');
  writeln(ErrOutput,
    '    removes the data (whose properties (/KEY and /VALUE only) are');
  writeln(ErrOutput, '    defined by the <options>)');
  writeln(ErrOutput,
    '    if no such data exists in the registry.pol file, this command does');
  writeln(ErrOutput, '    nothing.');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /VERSION');
  writeln(ErrOutput, '    prints the gpscript version to stdout.');
  writeln(ErrOutput);
  writeln(ErrOutput, '    data formats:');
  writeln(ErrOutput, '    dword:');
  writeln(ErrOutput, '      either decimal or hexadecimal (prefixed by 0x)');
  writeln(ErrOutput, '      dword prefix: D (example /DATA:D 0x00000002)');
  writeln(ErrOutput);
  writeln(ErrOutput, '    string:');
  writeln(ErrOutput,
    '      prefixed by ''S "'', suffixed by a single double quote ''"''');
  writeln(ErrOutput, '      the backslash "\" acts as an escape character:');
  writeln(ErrOutput, '      \\ -> \');
  writeln(ErrOutput, '      \" -> "');
  writeln(ErrOutput,
    '      \0x<hex> -> 2 digits hexadecimal representation of the character value');
  writeln(ErrOutput,
    '      string prefix: S (example /DATA:S "this\0x0020is a backslash \\.")');
  writeln(ErrOutput);
  writeln(ErrOutput, '    binary:');
  writeln(ErrOutput,
    '      formatted like binary data in .reg files (an unprefixed 2 digits');
  writeln(ErrOutput,
    '      hex value for each data byte, multiple bytes are separated by a blank');
  writeln(ErrOutput,
    '      or a comma. long lines may be wrapped using the backslash "\" as the');
  writeln(ErrOutput, '      last printeable character in a line).');
  writeln(ErrOutput, '      binary prefix: B (example /DATA:B FF 20 a0)');
  writeln(ErrOutput);
  writeln(ErrOutput, 'NOTE!');
  writeln(ErrOutput,
    '  under xp, you should run "gpupdate /force /wait:0" after using gpscript.');
  writeln(ErrOutput,
    '  under w2k, you should run "secedit /refreshpolicy machine_policy /enforce"');
  writeln(ErrOutput,
    '  or "secedit /refreshpolicy user_policy /enforce". if the changed policies');
  writeln(ErrOutput,
    '  still won''t get applied after those commands, you will have to open the');
  writeln(ErrOutput,
    '  group policy editor manually, modify and change back any value in both the');
  writeln(ErrOutput,
    '  machine and the user "administrative templates" part to get changes applied.');
  Halt(1);
end;

// check if an arg is given

procedure NeedsArg(strCmd, strArg: string; strErr: string);
begin
  if strArg = '' then
    raise Exception.CreateFmt(ERR_ARG_NEEDED, [strErr, strCmd]);
end;

// no arg must be given

procedure NoArg(strCmd, strArg: string; strErr: string);
begin
  if strArg <> '' then
    raise Exception.CreateFmt(ERR_ARG_NOT_ALLOWED, [strErr, strCmd]);
end;

procedure IncludeFile(AList: TStrings; AFile: string); forward;

// add a string to the command list

procedure CheckAddCommand(strCmd: string; List: TStrings; strErr: string);
var
  strArg: string;
  intPos, intType: Integer;
begin
  if strErr = '' then
    strErr := '(STDIN)';
  strCmd := Trim(strCmd);
  if (strCmd = '') or (strCmd[1] = '#') then
    Exit;
  if strCmd[1] <> '/' then
    raise Exception.CreateFmt(ERR_NOT_COMMAND, [strErr, strCmd]);
  intPos := Pos(':', strCmd);
  if intPos > 0 then
  begin
    strArg := Copy(strCmd, intPos + 1, MaxInt);
    strCmd := Copy(strCmd, 1, intPos - 1);
  end
  else
    strArg := '';
  // check command name
  case TArg(AnsiIndexText(UpperCase(strCmd), KNOWN_CMDS)) of
    aOKey:
      begin
        NeedsArg(strCmd, strArg, strErr);
        List.AddObject(strArg, Pointer(aOKey));
      end;
    aOValue:
      begin
        List.AddObject(strArg, Pointer(aOValue));
      end;
    aOType:
      begin
        NeedsArg(strCmd, strArg, strErr);
        // check type
        if not TryStrToInt(strArg, intType) then
        begin
          intType := AnsiIndexText(UpperCase(strArg), RegTypeNames);
          if intType = -1 then
            raise Exception.CreateFmt(ERR_UNKNOWN_TYPE, [strArg, strErr]);
        end;
        List.AddObject(IntToStr(intType), Pointer(aOType));
      end;
    aOData:
      begin
        List.AddObject(strArg, Pointer(aOData));
      end;
    aFCreate: DoCreate := True;
    aCClear:
      begin
        NoArg(strCmd, strArg, strErr);
        List.AddObject(strArg, Pointer(aCClear));
      end;
    aCGptIni: List.AddObject(strArg, Pointer(aCGptIni));
    aCDump:
      begin
        NoArg(strCmd, strArg, strErr);
        List.AddObject(strArg, Pointer(aCDump));
      end;
    aCFile:
      begin
        IncludeFile(List, strArg);
      end;
    aCSet:
      begin
        NoArg(strCmd, strArg, strErr);
        List.AddObject(strArg, Pointer(aCSet));
      end;
    aCAdd:
      begin
        NoArg(strCmd, strArg, strErr);
        List.AddObject(strArg, Pointer(aCAdd));
      end;
    aCModify:
      begin
        NoArg(strCmd, strArg, strErr);
        List.AddObject(strArg, Pointer(aCModify));
      end;
    aCDelete:
      begin
        NoArg(strCmd, strArg, strErr);
        List.AddObject(strArg, Pointer(aCDelete));
      end;
{$IFDEF DEBUG}
    aFDebug: DoDebug := True;
{$ENDIF}
  else
    raise Exception.CreateFmt(ERR_UNKNOWN_ARG, [strErr, strCmd]);
  end;
end;

// include the contents of another file

procedure IncludeFile(AList: TStrings; AFile: string);
var
  sl1: TStrings;
  intLoop: Integer;
  s, sLines: string;
  bInData: Boolean;
begin
  if AFile = '' then
    AList.AddObject('(STDIN)', Pointer(aOErrFile))
  else
    AList.AddObject(AFile, Pointer(aOErrFile));
  sl1 := TStringList.Create;
  try
    if AFile <> '' then
      sl1.LoadFromFile(AFile)
    else
      GetStdInList(sl1);

    with sl1 do
      if Bool(Count) then
      begin
        intLoop := 0;
        sLines := '';
        bInData := False;
        while intLoop < Count do
        begin
          // check for multi-line binary data
          if not bInData then
          begin
            if Pos('/DATA:B', Trim(UpperCase(Strings[intLoop]))) <> 1 then
              CheckAddCommand(Strings[intLoop], AList, AFile)
            else
            begin
              s := Trim(Strings[intLoop]);
              if Copy(s, Length(s), 1) <> '\' then
                CheckAddCommand(Strings[intLoop], AList, AFile)
              else
              begin
                bInData := True;
                sLines := Copy(s, 1, Length(s) - 1) + ' ';
              end;
            end;
          end
          else
          begin
            s := Trim(Strings[intLoop]);
            if Copy(s, Length(s), 1) <> '\' then
            begin
              CheckAddCommand(sLines + s, AList, AFile);
              bInData := False
            end
            else
              sLines := sLines + Copy(s, 1, Length(s) - 1) + ' ';
          end;
          Inc(intLoop);
        end;
      end;
  finally
    sl1.Free;
  end;
end;

// get the local policy filename (%systemroot%\system32\grouppolicy\<NAME>\egistry.pol)

procedure GetLocalGPFileName(var FName: string);
begin
  FName := GetSysDir + 'GroupPolicy\' + FName + '\Registry.pol';
end;

// get the name of the local gpt.ini file

function GetLocalGPTIniName: string;
begin
  Result := GetSysDir + 'GroupPolicy\gpt.ini';
end;

// collect all commands in one string list

procedure GetCommands(List: TStrings);
var
  intLoop: Integer;
begin
  List.Clear;
  FFileName := ParamStr(1);
  if (UpperCase(FFileName) = 'MACHINE') or (UpperCase(FFileName) = 'USER') then
  begin
    if UpperCase(FFileName) = 'MACHINE' then
      POL_TYPE := ptMachine
    else
      POL_TYPE := ptUser;
    GetLocalGPFileName(FFileName);
  end;

  List.AddObject('(parameters)', Pointer(aOErrFile));
  for intLoop := 2 to ParamCount do
    CheckAddCommand(ParamStr(intLoop), List, '(parameters)');
end;

type
  TModFlag = (mfModified, mfKeySet, mfValueSet, mfTypeSet, mfDataSet);
  TModFlags = set of TModFlag;

  // check if all necessary arguements have been passed to the app

procedure CheckParams(sErrCmd, sErrFile: string; seSet, seNeeded: TModFlags);
var
  sErr: string;
begin
  sErr := '';
  if (mfKeySet in seNeeded) and (not (mfKeySet in seSet)) then
    sErr := KNOWN_CMDS[Integer(aOKey)]
  else if (mfValueSet in seNeeded) and (not (mfValueSet in seSet)) then
    sErr := KNOWN_CMDS[Integer(aOValue)]
  else if (mfTypeSet in seNeeded) and (not (mfTypeSet in seSet)) then
    sErr := KNOWN_CMDS[Integer(aOType)]
  else if (mfDataSet in seNeeded) and (not (mfDataSet in seSet)) then
    sErr := KNOWN_CMDS[Integer(aOData)];
  if sErr <> '' then
    raise Exception.CreateFmt(ERR_ARG_MISSING, [sErr, sErrCmd, sErrFile]);
end;

// convert meta characters back to their character values

procedure ConvertStringData(var s: string);
var
  intPos: Integer;
begin
  intPos := 1;
  while intPos < Length(s) do
  begin
    if (s[intPos] = '\') and (intPos < Length(s)) then
      case s[intPos + 1] of
        '\', '"':
          begin
            Delete(s, intPos, 1);
            Inc(intPos);
          end;
        '0':
          if (Copy(s, intPos + 2, 1) = 'x') or (Copy(s, intPos + 2, 1) = 'X')
            then
          begin
            s := Copy(s, 1, intPos - 1) + Char(StrToInt(Copy(s, intPos + 1, 4)))
              + Copy(s, intPos + 5, MaxInt);
          end;
      end;
    Inc(intPos);
  end;
  if Copy(s, Length(s), 1) <> #0 then
    s := s + #0;
end;

// packs a unicode str in an ansi str

procedure PackUnicodeInAnsiString(const strW: WideString; var strA: AnsiString);
begin
  SetLength(strA, Length(strW) * sizeof(WideChar));
  Move(strW[1], strA[1], Length(strA));
end;

// process the commands list
// returns true if changes have been made

function ExecuteCommands(AList: TStrings): boolean;
var
  objPol: TRegPol;
  setFlags: TModFlags;
  intLoop: Integer;
  enKind: TArg;
  sErr: string;
  sTemp,
    sKey,
    sValue: string;
  sBin,
    sData: string; // contains the actual data
  intTemp: Integer;
  dwType: Cardinal;

{$IFDEF DEBUG}
  procedure Debug(strForm: string; DumpData: Boolean);
  var
    intLoop,
      intStart: Integer;
    strLine,
      strLine2: string;
  begin
    writeln(Format(strForm, [sValue, sKey]));
    if DumpData then
    begin
      writeln('type: ', GetDataTypeString(dwType), ' data:');
      intStart := 0;
      strLine := '';
      strLine2 := '';
      for intLoop := 1 to Length(sData) do
      begin
        strLine := strLine + IntToHex(Byte(sData[intLoop]), 2) + ' ';
        if sData[intLoop] > #31 then
          strLine2 := strLine2 + sData[intLoop]
        else
          strLine2 := strLine2 + '.';
        if ((intLoop mod 16) = 0) or (intLoop = Length(sData)) then
        begin
          write(IntToHex(intstart, 8), ': ');
          while Length(strLine) < 48 do
            strLine := strLine + ' ';
          writeln(strLine, ' ', strLine2);
          inc(intStart, 16);
          strLine := '';
          strLine2 := '';
        end;
      end;
      writeln;
    end;
    writeln('policy file: ', FFileName, #13#10, 'script file: ', sErr);
    writeln;
  end;
{$ENDIF}

begin
  setFlags := [];
  dwType := INVALID_HANDLE_VALUE;
  Result := False;
  if DoCreate and (not FileExists(FFileName)) then
  begin
    ForceDirectories(ExtractFilePath(FFileName));
    objPol := TRegPol.Create
  end
  else
    objPol := TRegPol.CreateFromFile(FFileName);
  try
    with AList do
      if Bool(Count) then
        for intLoop := 0 to Pred(Count) do
        begin
          enKind := TArg(Objects[intLoop]);
          case enKind of
            aOErrFile: sErr := Strings[intLoop];
            aOKey:
              begin
                sKey := UnQuote(Strings[intLoop]);
                Include(setFlags, mfKeySet);
              end;
            aOValue:
              begin
                sValue := UnQuote(Strings[intLoop]);
                Include(setFlags, mfValueSet);
              end;
            aOType:
              begin
                dwType := StrToInt(Strings[intLoop]);
                Include(setFlags, mfTypeSet);
              end;
            aOData:
              begin
                sTemp := UnQuote(Copy(Strings[intLoop], 2, MaxInt));
                case UpperCase((Strings[intLoop] + #0)[1])[1] of
                  'S':
                    begin
                      // convert meta characters
                      ConvertStringData(sTemp);
                      PackUnicodeInAnsiString(sTemp, sData);
                    end;
                  'D':
                    begin
                      if not TryStrToInt(sTemp, intTemp) then
                        raise Exception.CreateFmt(ERR_INVALID_DWORD, [sTemp,
                          sErr]);
                      SetLength(sData, sizeof(Integer));
                      system.Move(intTemp, sData[1], Length(sData));
                    end;
                  'B':
                    begin
                      sData := '';
                      if Bool(Length(sTemp)) then
                      begin
                        intTemp := 1;
                        while intTemp < Length(sTemp) do
                        begin
                          sBin := '0x' + Copy(sTemp, intTemp, 2);
                          inc(intTemp, 2);
                          while (intTemp < Length(sTemp)) and (sTemp[intTemp] in
                            [' ', ',', '\']) do
                            inc(intTemp);
                          try
                            sData := sData + Char(StrToInt(sBin));
                          except
                            raise Exception.CreateFmt(ERR_INVALID_BINARY,
                              [sTemp, sErr]);
                          end;
                        end;
                      end;
                    end;
                else
                  raise Exception.CreateFmt(ERR_UNKNOWN_DATA, [Strings[intLoop],
                    sErr]);
                end;
                Include(setFlags, mfDataSet);
              end;
            aCClear:
              begin
                objPol.Clear;
                Include(SetFlags, mfModified);
              end;
            aCGptIni:
              begin
                sTemp := UnQuote(Strings[intLoop]);
                if sTemp = '' then
                  sTemp := GetLocalGPTIniName;
                intTemp := GetPrivateProfileInt('general', 'version', 2+(65536*2),
                  PChar(sTemp));
                // increment value (machine and/or user part)
                case POL_TYPE of
                  ptUser: Inc(intTemp, 65536); // flag shows that user/registry.pol has been updated
                  ptMachine: Inc(intTemp, 1);  // flag shows that machine/registry.pol has been updated
                else
                  Inc(intTemp, 65537);
                end;
                WritePrivateProfileString('general', 'version',
                  pchar(IntToStr(intTemp)), PChar(sTemp));

                // add gpext.dll guids to gpt.ini so the administrative
                // template changes are put into the registry by the
                // gpedit.msc com interfaces
                if POL_TYPE in [ptMachine, ptFile] then
                begin
                  SetLength(sValue, 1025);
                  SetLength(sValue, GetPrivateProfileString('general',
                    'gPCMachineExtensionNames', nil, PChar(sValue), 1024,
                    PChar(sTemp)));
                  if Pos(UpperCase(GPEXT_DLL_MACHINE),
                    UpperCase(sValue)) = 0 then
                  begin
                    sValue := sValue + GPEXT_DLL_MACHINE;
                    WritePrivateProfileString('general',
                      'gPCMachineExtensionNames', PChar(sValue), PChar(sTemp));
                  end;
                end;

                if POL_TYPE in [ptUser, ptFile] then
                begin
                  SetLength(sValue, 1025);
                  SetLength(sValue, GetPrivateProfileString('general',
                    'gPCUserExtensionNames', nil, PChar(sValue), 1024,
                    PChar(sTemp)));
                  if Pos(UpperCase(GPEXT_DLL_USER),
                    UpperCase(sValue)) = 0 then
                  begin
                    sValue := sValue + GPEXT_DLL_USER;
                    WritePrivateProfileString('general',
                      'gPCUserExtensionNames', PChar(sValue), PChar(sTemp));
                  end;
                end;
              end;
            aCDump: objPol.Dump;
            aCSet:
              begin
                CheckParams(KNOWN_CMDS[Integer(aCSet)], sErr, setFlags,
                  [mfKeySet, mfValueSet, mfTypeSet, mfDataSet]);
{$IFDEF DEBUG}
                if DoDebug then
                begin
                  Debug('setting value "%s" in key "%s"', True);
                end
                else
{$ENDIF}
                  objPol.SetData(sKey, sValue, dwType, Length(sData),
                    @sData[1]);
                setFlags := [mfKeySet, mfModified];
              end;
            aCAdd:
              begin
                CheckParams(KNOWN_CMDS[Integer(aCAdd)], sErr, setFlags,
                  [mfKeySet, mfValueSet, mfTypeSet, mfDataSet]);
{$IFDEF DEBUG}
                if DoDebug then
                begin
                  Debug('adding value "%s" to key "%s"', True);
                end
                else
{$ENDIF}
                  objPol.AddData(sKey, sValue, dwType, Length(sData),
                    @sData[1]);
                setFlags := [mfKeySet, mfModified];
              end;
            aCModify:
              begin
                CheckParams(KNOWN_CMDS[Integer(aCModify)], sErr, setFlags,
                  [mfKeySet, mfValueSet, mfTypeSet, mfDataSet]);
{$IFDEF DEBUG}
                if DoDebug then
                begin
                  Debug('modifying value "%s" in key "%s"', True);
                end
                else
{$ENDIF}
                  objPol.ModifyData(sKey, sValue, dwType, Length(sData),
                    @sData[1]);
                setFlags := [mfKeySet, mfModified];
              end;
            aCDelete:
              begin
                CheckParams(KNOWN_CMDS[Integer(aCDelete)], sErr, setFlags,
                  [mfKeySet, mfValueSet]);
{$IFDEF DEBUG}
                if DoDebug then
                begin
                  Debug('deleting value "%s" from key "%s"', False);
                end
                else
{$ENDIF}
                  objPol.DeleteData(sKey, sValue);
                setFlags := [mfKeySet, mfModified];
              end;
          end;
        end;
{$IFDEF DEBUG}
    if not DoDebug then
{$ENDIF}
      if mfModified in setFlags then
      begin
        objPol.SaveToFile(FFileName);
        Result := True;
      end;
  finally
    objPol.Free;
  end;
end;

var
  slArgs: TStrings;
  bSetCmd: boolean;
  // void main
begin
  bSetCmd := False;
  try
    if FindACmdLineSwitch(['version']) then
    begin
      writeln(GPScriptVersion, ' (', RegPol2kXPVersion, ')');
      Halt(0);
    end;

    if (ParamCount < 2) or FindACmdLineSwitch(['h', 'help', '?', 'usage']) then
      Usage;

    // retrieve all commands and store them in a string list
    slArgs := TStringList.Create;
    try
      GetCommands(slArgs);
      bSetCmd := ExecuteCommands(slArgs);
    finally
      slArgs.Free;
    end;
  except
    on E: Exception do
    begin
      WriteLn(ErrOutput, ToOem(Format(STR_ERROR, [E.Message])));
      if ExitCode = 0 then
        ExitCode := 1;
    end;
  end;
  if (ExitCode = 0) and bSetCmd then
  begin
    writeln('under xp, you should run "gpupdate /force /wait:0" right now.');
    writeln('under w2k, you should run "secedit /refreshpolicy machine_policy /enforce"');
    writeln('or "secedit /refreshpolicy user_policy /enforce" right now.');
    writeln('if the changed policies still aren''t applied after those commands,');
    writeln('you will have to open the group policy editor manually, modify and');
    writeln('change back any value in both the machine and the user');
    writeln('"administrative templates" part to get changes applied.');
  end;
end.

