program ini2reg;

(*

  read ini file and store in registry

  program version: BT_VERSION, alpha stage!

  (C) by markus stephany merkes@mirkes.de, released under the GNU GPL

*)

{$APPTYPE CONSOLE}

uses
  Windows,
  SysUtils,
  Classes,
  inifiles,
  registry,
  strutils,
  utils,
  cl_utils;

{$I ..\LIB\VERSION.INC}

const
  Ini2RegVersion = 'ini2reg '+BT_VERSION; // my own version

// show help

procedure Usage;
begin
  writeln(ErrOutput, Ini2RegVersion);
  writeln(ErrOutput, CLEFT);
  writeln(ErrOutput);
  writeln(ErrOutput,
    'ini2reg reads values from an inifile and stores the data in the registry.');
  writeln(ErrOutput,
    'currently supported are REG_SZ, REG_EXPAND_SZ and REG_DWORD.');
  writeln(ErrOutput);
  writeln(ErrOutput, 'usage:');
  writeln(ErrOutput,
    'ini2reg <ini> <regkey> [<flags>] [<section[=<value>]> [<section[=<value>]>...]]');
  writeln(ErrOutput);
  writeln(ErrOutput);
  writeln(ErrOutput, '  <ini>:');
  writeln(ErrOutput,
    '    the name of the inifile whose values are to be read.');
  writeln(ErrOutput);
  writeln(ErrOutput, '  <regkey>:');
  writeln(ErrOutput,
    '    the name of the registry key to write the values to.');
  writeln(ErrOutput,
    '    known prefixes: HKLM/HKEY_LOCAL_MACHINE, HKCU/HKEY_CURRENT_USER,');
  writeln(ErrOutput,
    '    HKU/HKEY_USERS and HKCR/HKEY_CLASSES_ROOT.');
  writeln(ErrOutput);
  writeln(ErrOutput, '  <flags>:');
  writeln(ErrOutput, '    /OMITSECTION');
  writeln(ErrOutput,
    '    if this flag is specified, the values read from <ini> are written');
  writeln(ErrOutput,
    '    to <regkey> directly. if the flag is not specified, the values are');
  writeln(ErrOutput,
    '    written to a subkey of <regkey> named after the section where the');
  writeln(ErrOutput, '    value is located in the ini file.');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /NODWORD');
  writeln(ErrOutput,
    '    if this flag is specified, numeric values are not converted to REG_DWORD');
  writeln(ErrOutput,
    '    but stored as REG_SZ.');
  writeln(ErrOutput);
  writeln(ErrOutput, '    /NOEXPANDSZ[:RESOLVE]');
  writeln(ErrOutput,
    '    if this flag is specified, string values which contain environemt variables');
  writeln(ErrOutput,
    '    (%variable%) are stored as REG_SZ, not as REG_EXPAND_SZ.');
  writeln(ErrOutput,
    '    if the additional :RESOLVE argument is given, the current values of the');
  writeln(ErrOutput,
    '    environment variables will be stored, not their names.');
  writeln(ErrOutput);
  writeln(ErrOutput, '  <section>:');
  writeln(ErrOutput,
    '    you may specify one or more section names whose values are read from');
  writeln(ErrOutput,
    '    the ini file. if no section names are specified, the whole ini file');
  writeln(ErrOutput, '    is processed. wildcards (* and ?) are allowed.');
  writeln(ErrOutput);
  writeln(ErrOutput, '  <value>:');
  writeln(ErrOutput,
    '    for each section, a value name pattern can be specified. then only values');
  writeln(ErrOutput,
    '    matching this pattern are processed. wildcards (* and ?) are allowed.');
  Halt(1);
end;

resourcestring
  STR_ERR_UNKNOWN_OPTION = 'unknown option: %s';
  STR_FILE_NOT_FOUND = 'ini file not found: %s';

const
  // my options
  MY_COMMANDS: array[0..3] of string = (
    '/OMITSECTION',
    '/NODWORD',
    '/NOEXPANDSZ',
    '/NOEXPANDSZ:RESOLVE'
    );

var
  BoolNoSection: Boolean = False; // /omitsection
  BoolNoDWORD: Boolean = False; // /nodword
  BoolNoExpand: Boolean = False; // /noexpandsz
  BoolResolve: Boolean = False; // /noexpandsz:resolve

  StrFileName: string; // name of the ini file
  StrKeyName: string; // name of the key (without HK... prefix)
  HKRootKey: HKEY; // root key (cracked from the HK... prefix)

  objSects: TStrings;
  objIni: TMemIniFile;
  objIniSects,
    objIniVals: TStrings;
  objReg: TRegistry;

  intLoop, intLoop2: Integer;
  BoolFlagsOver: Boolean = False;

procedure Process(strSection, strName, strValue: string);
var
  intLoop: Integer;
  boolDo: Boolean;
  strKey: string;
  enType: TRegDataType;
  intVal: Integer;
begin
  boolDo := True;
  with objSects do
    if Bool(Count) then
    begin
      boolDo := False;
      for intLoop := 0 to Pred(Count) do
        if PatternMatch(Names[intLoop], strSection) and
          PatternMatch(Values[Names[intLoop]], strName) then
        begin
          boolDo := True;
          Break;
        end;
    end;
  if boolDo then
  begin
    // generate key name
    if BoolNoSection then
      strKey := ExcludeTrailingPathDelimiter(StrKeyName)
    else
      strKey :=
        ExcludeTrailingPathDelimiter(IncludeTrailingPathDelimiter(StrKeyName) +
        strSection);

    // check data type
    enType := rdString;
    if TryStrToInt(strValue, intVal) then
    begin
      if not BoolNoDWORD then
        // dword (integer in delphi)
        enType := rdInteger;
    end
    else
    begin
      // check if contains an even number of '%' signs (= contains %vars%)
      intVal := CharSetCount(['%'], strValue);
      if (intVal > 0) and ((intVal mod 2) = 0) then
      begin
        // is expandable
        if not BoolNoExpand then
          enType := rdExpandString
        else if BoolResolve then
          strValue := ExpandEnvironmentVariables(strValue);
      end;
    end;

    // open key
    objReg.OpenKey(strKey, True);
    try
      case enType of
        rdInteger: objReg.WriteInteger(strName, intVal);
        rdExpandString: objReg.WriteExpandString(strName, strValue);
      else
        objReg.WriteString(strName, strValue);
      end;
    finally
      objReg.CloseKey;
    end;
  end;
end;

begin
  try
    objSects := TStringList.Create;
    try
      if FindACmdLineSwitch(['version']) then
      begin
        writeln(Ini2RegVersion);
        Halt(0);
      end;

      if (ParamCount < 2) or FindACmdLineSwitch(['h', 'help', '?', 'usage'])
        then
        Usage;

      // crack the command line
      intLoop := 1;
      while intLoop <= ParamCount do
      begin
        case intLoop of
          1: StrFileName := ParamStr(1);
          2: CrackRegKey(ParamStr(2), hkRootKey, StrKeyName);
        else
          if not BoolFlagsOver then
          begin
            if Copy(ParamStr(intLoop), 1, 1) <> '/' then
            begin
              BoolFlagsOver := True;
              Dec(intLoop); // go back and continue
            end
            else
            begin
              // flags
              case AnsiIndexText(ParamStr(intLoop), MY_COMMANDS) of
                0: BoolNoSection := True;
                1: BoolNoDWORD := True;
                2: BoolNoExpand := True;
                3:
                  begin
                    BoolNoExpand := True;
                    BoolResolve := True;
                  end
              else
                raise Exception.CreateFmt(STR_ERR_UNKNOWN_OPTION,
                  [ParamStr(intLoop)]);
              end;
            end;
          end
          else
          begin
            // section argument
            if Pos('=', ParamStr(intLoop)) = 0 then
              objSects.Add(ParamStr(intLoop) + '=*')
            else
              objSects.Add(ParamStr(intLoop))
          end;
        end;
        Inc(intLoop);
      end;

      // add default '*=*', if list empty
      with objSects do
        if Count = 0 then
          Add('*=*');

      // load inifile
      if not FileExists(StrFileName) then
        raise Exception.CreateFmt(STR_FILE_NOT_FOUND, [StrFileName]);

      objReg := TRegistry.Create;
      try
        objReg.RootKey := hkRootKey;
        objIni := TMemIniFile.Create(StrFileName);
        try
          objIniSects := TStringList.Create;
          try
            objIniVals := TStringList.Create;
            try
              objIni.ReadSections(objIniSects);
              with objIniSects do
                if Bool(Count) then
                  for intLoop := 0 to Pred(Count) do
                  begin
                    objIni.ReadSection(Strings[intLoop], objIniVals);
                    with objIniVals do
                      if Bool(Count) then
                        for intLoop2 := 0 to Pred(Count) do
                          Process(objIniSects[intLoop], objIniVals[intLoop2],
                            objIni.ReadString(objIniSects[intLoop],
                            objIniVals[intLoop2], ''));
                  end;
            finally
              objIniVals.Free;
            end;
          finally
            objIniSects.Free;
          end;
        finally
          objIni.Free;
        end;
      finally
        objReg.Free;
      end;

    finally
      objSects.Free;
    end;
  except
    on E: Exception do
    begin
      WriteLn(ErrOutput, ToOem(Format(STR_ERROR, [E.Message])));
      if ExitCode = 0 then
        ExitCode := 1;
    end;
  end;
end.

