{*****************************************************************************************
 *
 * Project   : instfont
 * Unit Name : instfont
 * Purpose   : install or uninstall a font permanently
 * Author    : markus stephany
 * Copyright : (C)markus stephany, www.mirkes.de, merkes[at]mirkes[dot]de

      This package is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published
      by the Free Software Foundation; either version 2, or (at your option)
      any later version.

      This package is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with this package in the file named "COPYING"; if not,
      write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
      Boston, MA 02111-1307, USA.

 * Version   : 0.01, 01/29/2005
 * History   : 0.01 - 01/29/2005 : initial release
 *           : 0.02 - 01/30/2005 : bug fixes, forced font uninstallation, /p, /k flags added
 *
 *****************************************************************************************}

program instfont;

uses
  Windows, SysUtils, ActiveX, ShlObj;

{$I ..\LIB\VERSION.INC}

(******************* ErrWrite:
  write text to stderr *)
procedure ErrWrite(sMsg: string);
var
  dwDummy: Cardinal;
begin
  if sMsg <> ''
  then
    WriteFile(GetStdHandle(STD_ERROR_HANDLE), sMsg[1], Length(sMsg), dwDummy, nil);
end;

(******************* ErrWriteLn:
  put a line to stderr *)
procedure ErrWriteLn(sMsg: string);
begin
  ErrWrite(sMsg+#13#10);
end;

(******************* Bail:
 print an error message to stdout and exit *)
procedure Bail(sMsg: string; iExit: Integer=1);
begin
  ErrWriteLn(sMsg);
  Halt(iExit);
end;

(******************* Usage:
  print usage message *)
procedure Usage(sErrOpt:string = '');
var
  sApp: string;
begin
  sApp := ChangeFileExt(ExtractFileName(ParamStr(0)), '');
  // print the unknown option
  if sErrOpt <> '' then
  begin
    ErrWriteln(sApp+': unknown option "'+sErrOpt+'".');
    ErrWriteln('');
  end;
  ErrWriteln(sApp+' version '+BT_VERSION+' (C) markus stephany (GPL)');
  ErrWriteln('      (un)install a font permanently');
  ErrWriteln('usage:');
  ErrWriteln(sApp+' [/i] [/p] <font name> <font filename> = install a font');
  ErrWriteln('      <font name> is the name of the font as shown in font selection dialogs');
  ErrWriteln('      <font filename> is the actual filename of the font, the file is copied');
  ErrWriteln('      to the windows\fonts directory automatically before installation');
  ErrWriteln('      unless the /p flag is used (in this case the font is installed');
  ErrWriteln('      from its current location)');
  ErrWriteln(sApp+' /d [/p] [/k] <font name> <font filename> = uninstall a font');
  ErrWriteln('      <font name> is the name of the font as shown in font selection dialogs');
  ErrWriteln('      <font filename> is the name part of the font''s filename, it must be');
  ErrWriteln('      located in the windows\fonts directory unless the /p flag is used');
  ErrWriteln('      (in this case the font is uninstalled from its current location');
  ErrWriteln('      and the full path of the font may be specified). if the /k flag');
  ErrWriteln('      is specified, the font file is not removed from disk, the font');
  ErrWriteln('      will just be deleted from the system''s font list');
        Bail(sApp+' /? = shows this usage screen');
end;

(******************* CheckParam:
 check wheter paramstr(iIndex) is one of the specified options *)
function CheckParam(iIndex: Integer; const aChars: TSysCharSet): Boolean;
begin
  Result := (Length(ParamStr(iIndex)) = 2) and
            (ParamStr(iIndex)[1] in ['/','-']) and
            (ParamStr(iIndex)[2] in aChars);
end;

(******************* PrintVersion:
  print program version to stdout and exit *)
procedure PrintVersion;
begin
  writeln(ExtractFileName(ParamStr(0)),' version ',BT_VERSION);
  Halt(0);
end;

(******************* GetFontDirectory:
  retrieve the location of the system's font directory *)
function GetFontDirectory: string;
var
  pidL: PItemIDList;
  buf: array [0..Max_Path] of Char;
  malloc: IMalloc;
begin
  Result := '';
  // try special shell folder
  if SHGetSpecialFolderLocation (0, CSIDL_FONTS, pidL) = S_OK then
  begin
    FillChar(buf, sizeof(buf), 0);
    if SHGetPathFromIDList(pidl, buf) then
      SetString(Result, buf, StrLen(buf));
    if SHGetMalloc(malloc) = NOERROR then
        malloc.Free (pidl);
  end;
  if Result = '' then
  begin
    // fallback: try %windir%\fonts
    Result := GetEnvironmentVariable('windir');
    if Result <> '' then
      Result := IncludeTrailingPathDelimiter(Result)+'Fonts'
  end;
  if Result <> '' then
    Result := IncludeTrailingPathDelimiter(Result)
  else
    Bail('cannot retrieve the system''s font directory');
end;

const
  FONT_KEY = 'Software\Microsoft\Windows\CurrentVersion\Fonts';
  WM_FONTCHANGE = $001D;




(******************* ForceRemoveFontFile:
  if the font file exists, try to unload and remove it *)
procedure ForceDeleteFontFile(Font: string; IsOld: boolean);
begin
  if FileExists(Font) then
  begin
    if (not DeleteFile(Font)) or FileExists(Font) then
    begin
      // remove the font from the system's font table
      if RemoveFontResource(PChar(Font)) then
      begin
        SendNotifyMessage(HWND_BROADCAST, WM_FONTCHANGE, 0, 0);
        // wait five seconds to allow apps to unload the font
        Sleep(1500);
      end;

      // now delete the font file (3 tries)
      if FileExists(Font) and (not DeleteFile(Font)) then
      begin
        RemoveFontResource(PChar(Font));
        SendNotifyMessage(HWND_BROADCAST, WM_FONTCHANGE, 0, 0);
        Sleep(5000);
        if FileExists(Font) and (not DeleteFile(Font)) then
        begin
          RemoveFontResource(PChar(Font));
          SendNotifyMessage(HWND_BROADCAST, WM_FONTCHANGE, 0, 0);
          Sleep(5000);
          if FileExists(Font) and (not DeleteFile(Font)) then
          begin
            if IsOld then
              Bail('cannot delete previous font file '+Font)
            else
              Bail('cannot delete font file '+Font);
          end;
        end;
      end;
    end;
  end;
end;

(******************* InstallFont:
  install a font in the system after copying it to the system's font path *)
procedure InstallFont(Source, Target, Name: string);
var
  hksub: HKEY;
  tf: string;
begin
  // check if the target file already exists, if so, remove it from the system's
  // font list so it can be overwritten
  if Target <> '' then
  begin
    // copy source to target
    ForceDeleteFontFile(Target, True);
    if not CopyFile(PChar(Source), PChar(Target), False) then
      Bail('cannot copy '+Source+' to '+Target);
    // just file name, the font is in the windir\fonts directory
    tf := ExtractFileName(Target);
  end
  else
  begin
    // source = target, just temporarily remove font resource
    Target := Source;
    if RemoveFontResource(PChar(Source)) then
    begin
      SendNotifyMessage(HWND_BROADCAST, WM_FONTCHANGE, 0, 0);
      // wait five seconds to allow apps to unload the font
      Sleep(5000);
    end;
    // whole file Path, the font is not in the windir\fonts directory
    tf := Target;
  end;

  // depending on the operating system, store the new font in the registry
  if Win32Platform = VER_PLATFORM_WIN32_WINDOWS then
  begin
    if RegOpenKey(HKEY_LOCAL_MACHINE, FONT_KEY, hksub) <> ERROR_SUCCESS then
      Bail('cannot open registry key HKLM\'+FONT_KEY);
    begin
      if RegSetValueEx(hksub, PChar(Name), 0, REG_SZ, PChar(tf),
        Length(tf)+1) <> ERROR_SUCCESS then
      begin
        RegCloseKey(hksub);
        Bail('cannot register font in registry');
      end;
      RegCloseKey(hksub);
    end;
  end
  else
  begin
    if not WriteProfileString('Fonts', PChar(Name), PChar(tf))
    then
      Bail('cannot register font in registry');
  end;
  // register the font
  if AddFontResource(PChar(Target)) = 0 then
    Bail('cannot register font in the system');
  // send a font change notification to all top level windows
  SendNotifyMessage(HWND_BROADCAST, WM_FONTCHANGE, 0, 0);
end;

(******************* UninstallFont:
  uninstall a font file *)
procedure UninstallFont(Path, Name: string; KeepFile: boolean);
var
  hksub: HKEY;
  tf: string;
  rType: Cardinal;
  buf: array[0..MAX_PATH] of char;
begin
  // depending on the operating system, remove the font entry from the registry
  if Win32Platform = VER_PLATFORM_WIN32_WINDOWS then
  begin
    if RegOpenKey(HKEY_LOCAL_MACHINE, FONT_KEY, hksub) <> ERROR_SUCCESS then
      Bail('cannot open registry key HKLM\'+FONT_KEY);
    begin
      RegDeleteValue(hkSub, PChar(Name));
      if RegQueryValueEx(hkSub, PChar(Name), nil, @rType, nil, nil) = ERROR_SUCCESS then
      begin
        RegCloseKey(hksub);
        Bail('cannot unregister font in the registry');
      end;
      RegCloseKey(hksub);
    end;
  end
  else
  begin
    tf := '';
    UniqueString(tf);
    WriteProfileString('Fonts', PChar(Name), nil);
    FillChar(buf, sizeof(buf), 0);
    if GetProfileString('Fonts', PChar(Name), PChar(tf), buf, sizeof(buf)-1) > 0
      then
    begin
      SetString(tf, buf, StrLen(buf));
      if tf <> '' then
        Bail('cannot unregister font in the registry');
    end;
  end;

  // unload (and remove) the font file
  if not KeepFile then
    ForceDeleteFontFile(Path, False)
  else
  begin
    if RemoveFontResource(PChar(Path)) then
      SendNotifyMessage(HWND_BROADCAST, WM_FONTCHANGE, 0, 0);
  end;
end;


(******************* :
  void main() *)
var
  i,j:Integer;
  bInstall, bCurrentPath, bKeep: boolean;
  sFontName, sFontFile, sFontDir: string;
begin
  sFontName := '';
  sFontFile := '';
  bInstall := True;
  bCurrentPath := False;
  bKeep := False;
  j := 0;

  (* check the command line *)
  if ParamCount > 0 then
  begin
    for i := 1 to ParamCount
    do
      if CheckParam(i,['?','H','h'])
      then
        Usage
      else
        if CheckParam(i,['V'])
        then
          PrintVersion
        else
          if CheckParam(i,['i'])
          then
            bInstall := True
          else
            if CheckParam(i,['p'])
            then
              bCurrentPath := True
            else
              if CheckParam(i,['k'])
              then
                bKeep := True
              else
                if CheckParam(i,['d','u'])
                then
                  bInstall := False
                else
                begin
                  if j = 2 then
                    Usage(ParamStr(i));
                  if j = 0 then
                    sFontName := ParamStr(i)
                  else
                    sFontFile := ParamStr(i);
                  Inc(j);
                end;
  end;
  if (j <> 2) or (bKeep and bInstall) then
    Usage;

  if not bCurrentPath then
  begin
    sFontDir := GetFontDirectory;

    if not bInstall then
      sFontFile := sFontDir + ExtractFileName(sFontFile);


    if not FileExists(sFontFile) then
       Bail('Cannot locate font file '+sFontFile);

    if bInstall then
      InstallFont(sFontFile, sFontDir+ExtractFileName(sFontFile), sFontName)
    else
      UninstallFont(sFontFile, sFontName, bKeep);
  end
  else
  begin
    sFontFile := ExpandFileName(sFontFile);

    if not FileExists(sFontFile) then
       Bail('Cannot locate font file '+sFontFile);

    if bInstall then
      InstallFont(sFontFile, '', sFontName)
    else
      UninstallFont(sFontFile, sFontName, bKeep);
  end;
end.

