unit regpol2kxp;

(*

  file format specs and access routines of the 2k/xp registry.pol group policies file format

  unit version: 0.02, alpha stage!

  (C) markus stephany merkes@mirkes.de, released under the GNU GPL, version 2

*)

interface
uses
  Windows, SysUtils, Classes, utils;

const
  RegPolSignature = $67655250; // 'PREG' file header
  RegPolVersion = 1; // only supported version: 1
  RegPol2kXPVersion = 'regpol 2k/xp v0.0.2'; // my own version

  (* the registry.pol file body consists of registry values in the following format:
     [key;value;type;size;data]
     [key;  : path to the registry key without HKLM/HKCU, UNICODE!
     value; : name of the value, UNICODE!
              special value names:
                **DeleteValues:  multiple value names to delete, separated by semicolons
                **Del.valuename: name of a value to delete from the specified key
                **DelVals:       delete all values in the specified key
                **DeleteKeys:    multiple sub key names to delete, separated by semicolons
                **SecureKey:     **SecureKey=1: Administrators:F, EveryOne:R
                                 **SecureKey=0: reset access to inherited permissions
     type; : data type, DWORD!
     size; : size of the data following in bytes
     data] : user defined data
  *)

type
  // regpol exception
  ERegPol = class(Exception);

  // one policy entry
  TRegPolEntry = record
    KeyName, // name of the key without HKLM\HKCU
    ValName: PWideChar; // name of the value (or special value name, see above)
    DataType, // REG_xyz data type
    DataSize: Cardinal; // size of the Data buffer
    Data: Pointer; // value data
  end;
  PRegPolEntry = ^TRegPolEntry;

  // registry policy object
  TRegPol = class(TList)
  private
    FFileName: TFileName;
  protected
    procedure Put(AIndex: Integer; Item: PRegPolEntry);
    function Get(AIndex: Integer): PRegPolEntry;
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
  public
    constructor Create;
    constructor CreateFromFile(AFileName: TFileName);
    // create and load .pol file
    function Add(Item: PRegPolEntry): Integer; // add an item
    destructor Destroy; override;
    procedure LoadFromFile(AFileName: TFileName);
    // load from a different .pol file
    procedure ReLoad; // load file again
    procedure Save; // save to current file name
    procedure SaveToFile(AFileName: TFileName); // save to a different file name
    property Items[AIndex: Integer]: PRegPolEntry read Get write Put; default;
    procedure Dump; // dumps all entries in human/machine readable format
    function ByName(AKey, AValue: string): Integer;
    // find an entry by key/val name (nocase)
  // adding values - they must NOT already exist
    function AddData(AKey, AValue: string; DataType, DataSize: Cardinal; Data:
      Pointer): Integer;
    function AddString(AKey, AValue, AData: string; DataType: Cardinal =
      REG_SZ): Integer;
    function AddDWORD(AKey, AValue: string; Data: Cardinal): Integer;
    // modifiying values - they MUST already exist
    function ModifyData(AKey, AValue: string; DataType, DataSize: Cardinal;
      Data: Pointer): Integer;
    function ModifyString(AKey, AValue, AData: string; DataType: Cardinal =
      REG_SZ): Integer;
    function ModifyDWORD(AKey, AValue: string; Data: Cardinal): Integer;
    // setting values - they MAY already exist
    function SetData(AKey, AValue: string; DataType, DataSize: Cardinal; Data:
      Pointer): Integer;
    function SetString(AKey, AValue, AData: string; DataType: Cardinal =
      REG_SZ): Integer;
    function SetDWORD(AKey, AValue: string; Data: Cardinal): Integer;
    // deleting values
    procedure DeleteData(AKey, AValue: string);
  end;

implementation

resourcestring
  // error messages
  ERR_FILE_TOO_SMALL =
    'The file %s is too small to contain a valid group policy reg file.';
  ERR_INVALID_SIGNATURE = 'The file %s contains an invalid signature 0x%x.';
  ERR_UNSUPPORTED_VERSION = 'The file %s''s version 0x%x is unsupported.';
  ERR_UNEXPECTED_CHAR =
    'Unexpected character "%s" in %s at position 0x%x. Expected character: "%s".';
  ERR_EOF_UNEXPECTED = 'Unexpected EOF in %s';
  ERR_VALUE_EXISTS = 'Data for %s\%s already exists in the database.';
  ERR_NOT_EXISTS = '%s\%s does not exist in the database';

type
  // file header
  TRegPolHeader = packed record
    Signature,
      Version: DWORD;
  end;

  // data delimiter key;value;[type;size;]
  TRegPolDataDelim = packed record
    DataType: Cardinal;
    Delim1: WideChar;
    DataSize: Cardinal;
    Delim2: WideChar;
  end;

  { TRegPol }

constructor TRegPol.Create;
begin
  inherited Create;
  FFileName := '';
end;

constructor TRegPol.CreateFromFile(AFileName: TFileName);
begin
  Create;
  LoadFromFile(AFileName);
end;

destructor TRegPol.Destroy;
begin
  inherited;
end;

function TRegPol.Add(Item: PRegPolEntry): Integer;
begin
  Result := inherited Add(Item);
end;

function TRegPol.AddData(AKey, AValue: string; DataType, DataSize: Cardinal;
  Data: Pointer): Integer;
var
  pData: PRegPolEntry;
  strW: WideString;
begin
  if ByName(AKey, AValue) <> -1 then
    raise ERegPol.CreateFmt(ERR_VALUE_EXISTS, [AKey, AValue]);
  GetMem(pData, sizeof(TRegPolEntry));
  FillChar(pData^, sizeof(pData^), #0);
  pData^.DataType := DataType;
  pData^.DataSize := DataSize;
  GetMem(pData^.KeyName, (Length(AKey) + 1) * sizeof(WideChar));
  strW := AKey;
  System.Move(strW[1], pData^.KeyName^, (Length(AKey) + 1) * sizeof(WideChar));
  GetMem(pData^.ValName, (Length(AValue) + 1) * sizeof(WideChar));
  strW := AValue;
  System.Move(strW[1], pData^.ValName^, (Length(AValue) + 1) *
    sizeof(WideChar));
  if DataSize > 0 then
  begin
    GetMem(pData^.Data, DataSize);
    system.Move(Data^, pData^.Data^, DataSize);
  end;
  Result := Add(pData);
end;

function TRegPol.AddDWORD(AKey, AValue: string; Data: Cardinal): Integer;
begin
  Result := AddData(AKey, AValue, REG_DWORD, sizeof(DWORD), @Data);
end;

function TRegPol.AddString(AKey, AValue, AData: string; DataType: Cardinal =
  REG_SZ): Integer;
var
  strW: WideString;
begin
  if Length(AData) > 0 then
  begin
    strW := AData;
    Result := AddData(AKey, AValue, DataType, (Length(AData) + 1) *
      sizeof(WideChar), @strW[1])
  end
  else
    Result := AddData(AKey, AValue, DataType, 0, nil)
end;

function TRegPol.ModifyData(AKey, AValue: string; DataType, DataSize: Cardinal;
  Data: Pointer): Integer;
var
  iIndex: Integer;
begin
  iIndex := ByName(AKey, AValue);
  if iIndex = -1 then
    raise ERegPol.CreateFmt(ERR_NOT_EXISTS, [AKey, AValue]);
  Items[iIndex]^.DataType := DataType;
  if Items[iIndex]^.DataSize > 0 then
    FreeMem(Items[iIndex]^.Data);
  Items[iIndex]^.Data := nil;
  Items[iIndex]^.DataSize := DataSize;
  if DataSize > 0 then
  begin
    GetMem(Items[iIndex]^.Data, DataSize);
    system.Move(Data^, Items[iIndex]^.Data^, DataSize);
  end;
  Result := iIndex;
end;

function TRegPol.ModifyDWORD(AKey, AValue: string; Data: Cardinal): Integer;
begin
  Result := ModifyData(AKey, AValue, REG_DWORD, sizeof(DWORD), @Data);
end;

function TRegPol.ModifyString(AKey, AValue, AData: string; DataType: Cardinal =
  REG_SZ): Integer;
var
  strW: WideString;
begin
  if Length(AData) > 0 then
  begin
    strW := AData;
    Result := ModifyData(AKey, AValue, DataType, (Length(AData) + 1) *
      sizeof(WideChar), @strW[1])
  end
  else
    Result := ModifyData(AKey, AValue, DataType, 0, nil)
end;

function TRegPol.SetData(AKey, AValue: string; DataType, DataSize: Cardinal;
  Data: Pointer): Integer;
begin
  if ByName(AKey, AValue) = -1 then
    Result := AddData(AKey, AValue, DataType, DataSize, Data)
  else
    Result := ModifyData(AKey, AValue, DataType, DataSize, Data);
end;

function TRegPol.SetDWORD(AKey, AValue: string; Data: Cardinal): Integer;
begin
  Result := SetData(AKey, AValue, REG_DWORD, sizeof(DWORD), @Data);
end;

function TRegPol.SetString(AKey, AValue, AData: string; DataType: Cardinal =
  REG_SZ): Integer;
var
  strW: WideString;
begin
  if Length(AData) > 0 then
  begin
    strW := AData;
    Result := SetData(AKey, AValue, DataType, (Length(AData) + 1) *
      sizeof(WideChar), @strW[1])
  end
  else
    Result := SetData(AKey, AValue, DataType, 0, nil)
end;

procedure TRegPol.DeleteData(AKey, AValue: string);
begin
  if ByName(AKey, AValue) <> -1 then
  begin
    Delete(ByName(AKey, AValue));
    Pack;
  end;
end;

function TRegPol.Get(AIndex: Integer): PRegPolEntry;
begin
  Result := PRegPolEntry(inherited Get(AIndex));
end;

procedure TRegPol.Put(AIndex: Integer; Item: PRegPolEntry);
begin
  inherited Put(AIndex, Item);
end;

procedure TRegPol.ReLoad;
begin
  LoadFromFile(FFileName);
end;

procedure TRegPol.Save;
begin
  SaveToFile(FFileName);
end;

procedure TRegPol.Notify(Ptr: Pointer; Action: TListNotification);
var
  pData: PRegPolEntry absolute Ptr;
begin
  inherited;
  if (Action = lnDeleted) and Assigned(Ptr) then
    with pData^ do
    begin
      if Assigned(Data) then
        FreeMem(Data);
      if Assigned(KeyName) then
        FreeMem(KeyName);
      if Assigned(ValName) then
        FreeMem(ValName);
      FreeMem(pData);
    end;
end;

procedure TRegPol.LoadFromFile(AFileName: TFileName);
var
  strM: TMemoryStream;
  recHdr: TRegPolHeader;
  strKey,
    strVal: string;
  wchrRead: WideChar;
  recDelim: TRegPolDataDelim;
  pData: Pointer;
begin
  Clear;
  FFileName := AFileName;
  strM := TMemoryStream.Create;
  try
    strM.LoadFromFile(FFileName);
    if strM.Size < sizeof(TRegPolHeader) then
      raise ERegPol.CreateFmt(ERR_FILE_TOO_SMALL, [FFileName]);
    strM.Position := 0;

    // check file header
    strM.Read(recHdr, sizeof(recHdr));
    if recHdr.Signature <> RegPolSignature then
      raise ERegPol.CreateFmt(ERR_INVALID_SIGNATURE, [FFileName,
        recHdr.Signature]);
    if recHdr.Version <> RegPolVersion then
      raise ERegPol.CreateFmt(ERR_UNSUPPORTED_VERSION, [FFileName,
        recHdr.Version]);

    // now read all entries
    try
      while strM.Position < strM.Size do
      begin
        strKey := '';
        strVal := '';
        strM.Read(wchrRead, sizeof(wChrRead));

        // must be a [
        if wchrRead <> '[' then
          raise ERegPol.CreateFmt(ERR_UNEXPECTED_CHAR, [wchrRead, FFileName,
            strM.Position, '[']);

        // read key, until ;
        repeat
          strM.Read(wchrRead, sizeof(wchrRead));
          if wchrRead = ';' then
            Break;
          if wchrRead <> #0 then
            strKey := strKey + wchrRead;
          if strM.Position >= strM.Size then
            raise ERegPol.CreateFmt(ERR_EOF_UNEXPECTED, [FFileName]);
        until False;

        // read value name, until ;
        repeat
          strM.Read(wchrRead, sizeof(wchrRead));
          if wchrRead = ';' then
            Break;
          if wchrRead <> #0 then
            strVal := strVal + wchrRead;
          if strM.Position >= strM.Size then
            raise ERegPol.CreateFmt(ERR_EOF_UNEXPECTED, [FFileName]);
        until False;

        // read data type and size
        strM.Read(recDelim, sizeof(recDelim));
        if strM.Position >= strM.Size then
          raise ERegPol.CreateFmt(ERR_EOF_UNEXPECTED, [FFileName]);

        if (recDelim.Delim1 <> ';') then
          raise ERegPol.CreateFmt(ERR_UNEXPECTED_CHAR, [string(recDelim.Delim1),
            FFileName, strM.Position - sizeof(recDelim) + 1, ';']);
        if (recDelim.Delim2 <> ';') then
          raise ERegPol.CreateFmt(ERR_UNEXPECTED_CHAR, [string(recDelim.Delim2),
            FFileName, strM.Position - sizeof(recDelim) + 1, ';']);

        // read Data
        pData := nil;
        try
          if recDelim.DataSize > 0 then
          begin
            GetMem(pData, recDelim.DataSize);
            strM.Read(pData^, recDelim.DataSize);
            if strM.Position >= strM.Size then
              raise ERegPol.CreateFmt(ERR_EOF_UNEXPECTED, [FFileName]);
          end;

          // read ]
          strM.Read(wchrRead, sizeof(wchrRead));
          // must be a [
          if wchrRead <> ']' then
            raise ERegPol.CreateFmt(ERR_UNEXPECTED_CHAR, [wchrRead, FFileName,
              strM.Position, ']']);

          AddData(strKey, strVal, recDelim.DataType, recDelim.DataSize, pData);
        finally
          if Assigned(pData) then
            FreeMem(pData);
        end;
      end;
    except
      Clear;
      raise;
    end;
  finally
    strM.Free;
  end;
end;

procedure TRegPol.SaveToFile(AFileName: TFileName);
var
  strF: TFileStream;
  recHdr: TRegPolHeader;
  intLoop: Integer;
  strW: WideString;
  chrW: WideChar;
  recDelim: TRegPolDataDelim;
begin
  strF := TFileStream.Create(AFileName, fmCreate);
  try
    recHdr.Signature := RegPolSignature;
    recHdr.Version := RegPolVersion;
    strF.Write(recHdr, sizeof(recHdr));
    recDelim.Delim1 := ';';
    recDelim.Delim2 := ';';
    if Count > 0 then
      for intLoop := 0 to Pred(Count) do
        with Items[intLoop]^ do
        begin
          chrW := ';';
          // write keyname
          strW := '[' + WideString(KeyName) + #0;
          strF.Write(strW[1], Length(strW) * sizeof(WideChar));
          // write delimiter
          strF.Write(chrW, sizeof(chrW));
          // write valname;
          strW := WideString(ValName) + #0;
          strF.Write(strW[1], Length(strW) * sizeof(WideChar));
          // write delimiter
          strF.Write(chrW, sizeof(chrW));
          // write type and size
          recDelim.DataType := DataType;
          recDelim.DataSize := DataSize;
          strF.Write(recDelim, sizeof(recDelim));
          // write data;
          if DataSize > 0 then
            strF.Write(Data^, DataSize);
          // write delimiter
          chrW := ']';
          strF.Write(chrW, sizeof(chrW));
        end;
    FFileName := AFileName;
  finally
    strF.Free;
  end;
end;

procedure TRegPol.Dump;
var
  intLoop: Integer;
  strW: WideString;
  strA: AnsiString;
begin
  // dumps all entries in human/gpscript readable format
  writeln(Format('# regpol group policy dump of %s on %s', [FFileName,
    ComputerName]));
  writeln(Format('# date/time: %s, version: %s', [FormatDateTime('c', now),
    RegPol2kXPVersion]));
  writeln;
  writeln('# (C) markus stephany, copylefted under the terms of the GNU GPL version 2.');
  if Count > 0 then
    for intLoop := 0 to Pred(Count) do
      with Items[intLoop]^ do
      begin
        writeln;
        writeln('/KEY:"', string(WideString(KeyName)), '"');
        writeln('/VALUE:"', string(WideString(ValName)), '"');
        writeln('/TYPE:', GetDataTypeString(DataType));
        write('/DATA:');
        case DataType of
          REG_SZ, REG_EXPAND_SZ, REG_MULTI_SZ:
            begin
              SetLength(strW, DataSize div sizeof(WideChar));
              system.Move(Data^, strW[1], Length(strW) * sizeof(WideChar));
              strA := strW;
              writeln(GetStringValue(Length(strA), @strA[1]));
            end;
          REG_DWORD: writeln(GetDWORDValue(DataSize, Data));
        else
          writeln(GetBinaryValue(DataSize, Data));
        end;
        writeln('/SET');
      end;
end;

function TRegPol.ByName(AKey, AValue: string): Integer;
var
  intLoop: Integer;
begin
  Result := -1;
  if Count > 0 then
    for intLoop := 0 to Pred(Count) do
      with Items[intLoop]^ do
      begin
        if (WideCompareText(KeyName, WideString(AKey)) = 0) and
          (WideCompareText(ValName, WideString(AValue)) = 0) then
        begin
          Result := intLoop;
          Break;
        end;
      end;
end;

end.

