program moveex;

(*

  rename a file, if it's currently locked, rename it after next reboot

  program version: BT_VERSION

  (C) by markus stephany merkes@mirkes.de, released under the GNU GPL, version 2

*)

{$APPTYPE CONSOLE}

uses
  Windows,
  SysUtils,
  utils;

{$I ..\LIB\VERSION.INC}
const
  MoveExVersion = 'moveex '+BT_VERSION; // my own version

  // show help

procedure Usage;
begin
  writeln(ErrOutput, MoveExVersion);
  writeln(ErrOutput, CLEFT);
  writeln(ErrOutput);
  writeln(ErrOutput,
    'moveex can be used to rename/delete files which are currently locked.');
  writeln(ErrOutput,
    'if the file is locked, renaming/deletion will delay until next reboot.');
  writeln(ErrOutput);
  writeln(ErrOutput, 'usage:');
  writeln(ErrOutput,
    'moveex <file to rename/delete> [<new file name>] [/q] [/f]');
  writeln(ErrOutput);
  writeln(ErrOutput);
  writeln(ErrOutput, '  <file to rename/delete>:');
  writeln(ErrOutput,
    '    the source name of the file to be renamed/deleted, either including the');
  writeln(ErrOutput,
    '    full path, a relative path or just the name of the file.');
  writeln(ErrOutput);
  writeln(ErrOutput, '  <new file name>:');
  writeln(ErrOutput,
    '    the destination name of the file. it must not contain a path');
  writeln(ErrOutput,
    '    because the file will always be renamed in its source directory.');
  writeln(ErrOutput,
    '    if this parameter is omitted, the source file will be deleted.');
  writeln(ErrOutput);
  writeln(ErrOutput, '  /q:');
  writeln(ErrOutput,
    '    if this flag is specified, moveex will work without printing');
  writeln(ErrOutput,
    '    status messages or asking confirmation questions.');
  writeln(ErrOutput);
  writeln(ErrOutput, '  /f:');
  writeln(ErrOutput,
    '    if this flag is specified, moveex will not try to rename/delete the file');
  writeln(ErrOutput,
    '    immediately but operation will always be delayed until next boot.');
  writeln(ErrOutput);
  writeln(ErrOutput, 'status error codes:');
  writeln(ErrOutput,
    '  %errorlevel% will be set to one of the following values:');
  writeln(ErrOutput);
  writeln(ErrOutput,
    '    0: the file could be renamed/deleted immediately');
  writeln(ErrOutput,
    '    1: an error occured (or execution cancelled)');
  writeln(ErrOutput,
    '    2: the file will be renamed/deleted after next reboot');
  writeln(ErrOutput);
  writeln(ErrOutput, 'how does it work?');
  writeln(ErrOutput,
    '  under winnt, there is an api called MoveFileEx to perform this task (it');
  writeln(ErrOutput,
    '  writes the source and target file name to the registry value');
  writeln(ErrOutput);
  writeln(ErrOutput,
    '  HKLM\SYSTEM\CurrentControlSet\Control\Session Manager\');
  writeln(ErrOutput,
    '  PendingFileRenameOperations');
  writeln(ErrOutput);
  writeln(ErrOutput,
    '  and at thus the file will be renamed at next boot).');
  writeln(ErrOutput);
  writeln(ErrOutput,
    '  under win9x, the MoveFileEx api is just a stub and doesn''t work, there');
  writeln(ErrOutput,
    '  a configuration file in the windows directory called wininit.ini is parsed');
  writeln(ErrOutput,
    '  at startup, and all entries in the RENAME section will be performed.');
  writeln(ErrOutput);
  writeln(ErrOutput,
    '  moveex knows whether it is called under nt or 9x and chooses the approriate');
  writeln(ErrOutput,
    '  method automatically. if one of the file names (source or target) are already');
  writeln(ErrOutput,
    '  stored in the registry value mentioned above, moveex will fail under winnt.');
  writeln(ErrOutput);
  writeln(ErrOutput, 'NOTE:');
  writeln(ErrOutput,
    '  on win9x, only short filenames should be used!');
  writeln(ErrOutput);
  Halt(1);
end;

// check for double entries in the registry (nt)

function IsInRegistry(strFile: string): Boolean;
var
  hkSub: HKEY;
  dwError: Cardinal;
  strValue,
    strSrc,
    strDst: string;
  dwVal: Cardinal;
  intPos: Integer;
begin
  Result := False;
  dwError := RegOpenKey(HKEY_LOCAL_MACHINE,
    'SYSTEM\CurrentControlSet\Control\Session Manager', hkSub);
  if dwError = 0 then
  begin
    dwError := RegQueryValueEx(hkSub, 'PendingFileRenameOperations', nil, nil,
      nil, @dwVal);
    if dwError = 0 then
    begin
      SetLength(strValue, dwVal);
      dwError := RegQueryValueEx(hkSub, 'PendingFileRenameOperations', nil, nil,
        @strValue[1], @dwVal);
      if (dwError = 0) and (dwVal > 0) then
      begin
        // now the string contains all values in the form source#0[dest|#0]#0#0
        repeat
          if (strValue = '') or (strValue[1] = #0) then
            Break;

          // source-file, must not be empty

          intPos := Pos(#0, strValue);
          if intPos in [0..1] then
            Break;
          strSrc := Copy(strValue, 1, intPos - 1);
          Delete(strValue, 1, intPos);

          // target-file, may be empty (for deletion)

          intPos := Pos(#0, strValue);
          if intPos = 0 then
            Break;
          strDst := Copy(strValue, 1, intPos - 1);
          Delete(strValue, 1, intPos);

          // compare both filenames with the given file in szFile (may be prefixed by '\??\')
          if SameFileName(strDst, strFile) or
            SameFileName(strDst, '\??\' + strFile) or
            SameFileName(strSrc, strFile) or
            SameFileName(strSrc, '\??\' + strFile) then
          begin
            Result := True;
            Break;
          end;

        until strValue = '';
      end;
    end;
    RegCloseKey(hkSub);
  end;
end;

// return a pchar or nil

function PCharOrNil(str: string): PChar;
begin
  if str = '' then
    Result := nil
  else
    Result := PChar(str);
end;

resourcestring
  STR_ERR_SRCNOTFOUND = 'source file %s not found';
  STR_ERR_DSTHASPATH = 'invalid destination name (contains a path): %s';
  STR_ERR_DSTISPATH = 'invalid destination name (it''s an existing path): %s';
  STR_ERR_REGRENAMEPENDING =
    'file %s is already scheduled to be renamed/deleted at next boot.';
  STR_ERR_ABORTED = 'aborted.';
  STR_ERR_CANNOTRENAME = 'cannot rename %s to %s (%s)';
  STR_ERR_CANNOTDELETE = 'cannot delete %s (%s)';
  STR_ERR_NOSHORTPATH = 'no short filename available for %s';

  STR_ASK_OVERWRITE = '%s already exists, overwrite?';
  STR_ASK_DELETE = 'do you really want to delete %s?';
  STR_YESNODEF = 'yN';
  STR_NO = 'n';
  STR_SUCCESS_DELETE = 'successfully deleted %s';
  STR_SUCCESS_RENAME = 'successfully renamed %s to %s';
  STR_NEXTREBOOT = 'changes will take effect at next reboot.';

var
  BoolQuiet: Boolean = False; // /q flag
  BoolDelete: Boolean = False; // no target file name given
  BoolForceDelay: Boolean = False; // /f flag

  StrSourceName: string = '';
  StrTargetName: string = ''; // source and target file name

  strAnswer: string; // answer for confirmation queries
  boolDone: Boolean; // success ?
  strIni: string; // wininit.ini
  intParams: Integer; // number of arguments
  hFile: THandle; // create target file on win9x
begin
  intParams := ParamCount;
  try
    if FindACmdLineSwitch(['version']) then
    begin
      writeln(MoveExVersion);
      Halt(0);
    end;

    if (not (ParamCount in [1..4])) or FindACmdLineSwitch(['h', 'help', '?',
      'usage']) then
      Usage;

    BoolQuiet := FindCmdLineSwitch('Q', True);
    if BoolQuiet then
      Dec(intParams);

    BoolForceDelay := FindCmdLineSwitch('F', True);
    if BoolForceDelay then
      Dec(intParams);

    // get filenames
    StrSourceName := ExpandFileName(ParamStr(1));
    if intParams = 1 then
      BoolDelete := True
    else
      StrTargetName := ParamStr(2);

    // check file names
    if not FileExists(StrSourceName) then
      raise Exception.CreateFmt(STR_ERR_SRCNOTFOUND, [StrSourceName]);

    // check target file name (if any)
    if (not BoolDelete) then
    begin
      if ExtractFileName(StrTargetName) <> StrTargetName then
        raise Exception.CreateFmt(STR_ERR_DSTHASPATH, [StrTargetName]);

      // complete target file name = path of source file name + target file name
      StrTargetName := ExtractFilePath(StrSourceName) + StrTargetName;

      // make sure that target is not an existing directory
      if DirectoryExists(StrTargetName) then
        raise Exception.CreateFmt(STR_ERR_DSTISPATH, [StrTargetName]);

      if (not BoolQuiet) and (FileExists(StrTargetName)) then
      begin
        // confirm overwriting
        strAnswer := Choice(Format(STR_ASK_OVERWRITE, [StrTargetName]),
          STR_YESNODEF, STR_NO[1]);
        if UpperCase(strAnswer) = UpperCase(STR_NO) then
          raise Exception.Create(STR_ERR_ABORTED);
      end;
    end
    else if not BoolQuiet then
    begin
      // confirm deletion
      strAnswer := Choice(Format(STR_ASK_DELETE, [StrSourceName]), STR_YESNODEF,
        STR_NO[1]);
      if UpperCase(strAnswer) = UpperCase(STR_NO) then
        raise Exception.Create(STR_ERR_ABORTED);
    end;

    // first try to immediately rename/delete the file if /f is not given
    SetFileAttributes(PChar(StrSourceName), 0);
    if not BoolDelete then
    begin
      SetFileAttributes(PChar(StrTargetName), 0);
      boolDone := (not BoolForceDelay) and ((not FileExists(StrTargetName)) or
        DeleteFile(PChar(StrTargetName))) and
        (not FileExists(StrTargetName)) and MoveFile(PChar(StrSourceName),
        PChar(StrTargetName))
    end
    else
      boolDone := (not BoolForceDelay) and DeleteFile(PChar(StrSourceName)) and
        (not FileExists(StrSourceName));

    if boolDone then
    begin
      ExitCode := 0;
      if not BoolQuiet then
      begin
        if BoolDelete then
          writeln(Format(STR_SUCCESS_DELETE, [StrSourceName]))
        else
          writeln(Format(STR_SUCCESS_RENAME, [StrSourceName, StrTargetName]))
      end;
    end
    else
    begin
      // delay until reboot
      if Win32Platform = VER_PLATFORM_WIN32_NT then
      begin
        // nt version

        // filename must not already be stored under PendingFileRenameOperations
        if IsInRegistry(StrSourceName) then
          raise Exception.CreateFmt(STR_ERR_REGRENAMEPENDING, [StrSourceName]);

        if (not BoolDelete) and IsInRegistry(StrTargetName) then
          raise Exception.CreateFmt(STR_ERR_REGRENAMEPENDING, [StrTargetName]);

        if (BoolDelete or MoveFileEx(PChar(StrTargetName), nil,
          MOVEFILE_DELAY_UNTIL_REBOOT)) and
          MoveFileEx(PChar(StrSourceName), PCharOrNil(StrTargetName),
          MOVEFILE_DELAY_UNTIL_REBOOT) then
        begin
          if not BoolQuiet then
            writeln(STR_NEXTREBOOT);
          ExitCode := 2;
        end
        else
        begin
          if not BoolDelete then
            raise Exception.CreateFmt(STR_ERR_CANNOTRENAME, [StrSourceName,
              StrTargetName, SysErrorMessage(GetLastError)])
          else
            raise Exception.CreateFmt(STR_ERR_CANNOTDELETE, [StrSourceName,
              SysErrorMessage(GetLastError)])
        end
      end
      else
      begin
        // w9x version
        if (not BoolDelete) then
        begin
          // make sure the file exists, so we get a valid destination name
          if not FileExists(StrTargetName) then
          begin
            hFile := CreateFile(PChar(StrTargetName), GENERIC_READ or
              GENERIC_WRITE, 0, nil, CREATE_ALWAYS, 0, 0);
            CloseHandle(hFile);
          end;

          if (ExtractShortPathName(StrTargetName) = '') then
            raise Exception.CreateFmt(STR_ERR_NOSHORTPATH,[StrTargetName]);
        end;

        strIni := ExpandEnvironmentVariables('%windir%\wininit.ini');
        if (BoolDelete or WritePrivateProfileString('Rename', 'NUL',
          PChar(ExtractShortPathName(StrTargetName)), PChar(strIni))) and
          WritePrivateProfileString('Rename',
          PChar(ExtractShortPathName(StrTargetName)),
          PChar(ExtractShortPathName(StrSourceName)), PChar(strIni)) then
        begin
          if not BoolQuiet then
            writeln(STR_NEXTREBOOT);
          ExitCode := 2;
        end
        else
        begin
          if not BoolDelete then
            raise Exception.CreateFmt(STR_ERR_CANNOTRENAME, [StrSourceName,
              StrTargetName, SysErrorMessage(GetLastError)])
          else
            raise Exception.CreateFmt(STR_ERR_CANNOTDELETE, [StrSourceName,
              SysErrorMessage(GetLastError)])
        end
      end;
    end;
  except
    on E: Exception do
    begin
      WriteLn(ErrOutput, ToOem(Format(STR_ERROR, [E.Message])));
      if ExitCode = 0 then
        ExitCode := 1;
    end;
  end;
end.

